import datetime
from typing import Dict, Iterator, List, Union, Tuple, Optional, Any
import copy

from langchain_core.tools import BaseTool
from sqlalchemy.orm.attributes import flag_modified
from sqlalchemy import null
import pandas as pd

from mindsdb.interfaces.storage import db
from mindsdb.interfaces.storage.db import Predictor
from mindsdb.utilities.context import context as ctx
from mindsdb.interfaces.data_catalog.data_catalog_loader import DataCatalogLoader
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.interfaces.model.functions import PredictorRecordNotFound
from mindsdb.interfaces.model.model_controller import ModelController
from mindsdb.interfaces.skills.skills_controller import SkillsController
from mindsdb.utilities.config import config
from mindsdb.utilities import log

from mindsdb.utilities.exception import EntityExistsError, EntityNotExistsError

from .constants import ASSISTANT_COLUMN, SUPPORTED_PROVIDERS, PROVIDER_TO_MODELS
from .langchain_agent import get_llm_provider

logger = log.getLogger(__name__)

default_project = config.get("default_project")


class AgentsController:
    """Handles CRUD operations at the database level for Agents"""

    assistant_column = ASSISTANT_COLUMN

    def __init__(
        self,
        project_controller: ProjectController = None,
        skills_controller: SkillsController = None,
        model_controller: ModelController = None,
    ):
        if project_controller is None:
            project_controller = ProjectController()
        if skills_controller is None:
            skills_controller = SkillsController()
        if model_controller is None:
            model_controller = ModelController()
        self.project_controller = project_controller
        self.skills_controller = skills_controller
        self.model_controller = model_controller

    def check_model_provider(self, model_name: str, provider: str = None) -> Tuple[dict, str]:
        """
        Checks if a model exists, and gets the provider of the model.

        The provider is either the provider of the model or the provider given as an argument.

        Parameters:
            model_name (str): The name of the model
            provider (str): The provider to check

        Returns:
            model (dict): The model object
            provider (str): The provider of the model
        """
        model = None

        # Handle the case when model_name is None (using default LLM)
        if model_name is None:
            return model, provider

        try:
            model_name_no_version, model_version = Predictor.get_name_and_version(model_name)
            model = self.model_controller.get_model(model_name_no_version, version=model_version)
            provider = "mindsdb" if model.get("provider") is None else model.get("provider")
        except PredictorRecordNotFound:
            if not provider:
                # If provider is not given, get it from the model name
                provider = get_llm_provider({"model_name": model_name})

            elif provider not in SUPPORTED_PROVIDERS and model_name not in PROVIDER_TO_MODELS.get(provider, []):
                raise ValueError(f"Model with name does not exist for provider {provider}: {model_name}")

        return model, provider

    def get_agent(self, agent_name: str, project_name: str = default_project) -> Optional[db.Agents]:
        """
        Gets an agent by name.

        Parameters:
            agent_name (str): The name of the agent
            project_name (str): The name of the containing project - must exist

        Returns:
            agent (Optional[db.Agents]): The database agent object
        """

        project = self.project_controller.get(name=project_name)
        agent = db.Agents.query.filter(
            db.Agents.name == agent_name,
            db.Agents.project_id == project.id,
            db.Agents.company_id == ctx.company_id,
            db.Agents.deleted_at == null(),
        ).first()
        return agent

    def get_agent_by_id(self, id: int, project_name: str = default_project) -> db.Agents:
        """
        Gets an agent by id.

        Parameters:
            id (int): The id of the agent
            project_name (str): The name of the containing project - must exist

        Returns:
            agent (db.Agents): The database agent object
        """

        project = self.project_controller.get(name=project_name)
        agent = db.Agents.query.filter(
            db.Agents.id == id,
            db.Agents.project_id == project.id,
            db.Agents.company_id == ctx.company_id,
            db.Agents.deleted_at == null(),
        ).first()
        return agent

    def get_agents(self, project_name: str) -> List[dict]:
        """
        Gets all agents in a project.

        Parameters:
            project_name (str): The name of the containing project - must exist

        Returns:
            all-agents (List[db.Agents]): List of database agent object
        """

        all_agents = db.Agents.query.filter(db.Agents.company_id == ctx.company_id, db.Agents.deleted_at == null())

        if project_name is not None:
            project = self.project_controller.get(name=project_name)

            all_agents = all_agents.filter(db.Agents.project_id == project.id)

        return all_agents.all()

    def _create_default_sql_skill(
        self,
        name,
        project_name,
        include_tables: List[str] = None,
        include_knowledge_bases: List[str] = None,
    ):
        # Create a default SQL skill
        skill_name = f"{name}_sql_skill"
        skill_params = {
            "type": "sql",
            "description": f"Auto-generated SQL skill for agent {name}",
        }

        # Add restrictions provided
        if include_tables:
            skill_params["include_tables"] = include_tables
        if include_knowledge_bases:
            skill_params["include_knowledge_bases"] = include_knowledge_bases

        try:
            # Check if skill already exists
            existing_skill = self.skills_controller.get_skill(skill_name, project_name)
            if existing_skill is None:
                # Create the skill
                skill_type = skill_params.pop("type")
                self.skills_controller.add_skill(
                    name=skill_name, project_name=project_name, type=skill_type, params=skill_params
                )
            else:
                # Update the skill if parameters have changed
                params_changed = False

                # Check if skill parameters need to be updated
                for param_key, param_value in skill_params.items():
                    if existing_skill.params.get(param_key) != param_value:
                        existing_skill.params[param_key] = param_value
                        params_changed = True

                # Update the skill if needed
                if params_changed:
                    flag_modified(existing_skill, "params")
                    db.session.commit()

        except Exception as e:
            raise ValueError(f"Failed to auto-create or update SQL skill: {str(e)}")

        return skill_name

    def add_agent(
        self,
        name: str,
        project_name: str = None,
        model_name: Union[str, dict] = None,
        skills: List[Union[str, dict]] = None,
        provider: str = None,
        params: Dict[str, Any] = None,
    ) -> db.Agents:
        """
        Adds an agent to the database.

        Parameters:
            name (str): The name of the new agent
            project_name (str): The containing project
            model_name (str | dict): The name of the existing ML model the agent will use
            skills (List[Union[str, dict]]): List of existing skill names to add to the new agent, or list of dicts
                 with one of keys is "name", and other is additional parameters for relationship agent<>skill
            provider (str): The provider of the model
            params (Dict[str, str]): Parameters to use when running the agent
                data: Dict, data sources for an agent, keys:
                  - knowledge_bases: List of KBs to use
                  - tables: list of tables to use
                model: Dict, parameters for the model to use
                  - provider: The provider of the model (e.g., 'openai', 'google')
                  - Other model-specific parameters like 'api_key', 'model_name', etc.
                <provider>_api_key: API key for the provider (e.g., openai_api_key)

                # Deprecated parameters:
                database: The database to use for text2sql skills (default is 'mindsdb')
                knowledge_base_database: The database to use for knowledge base queries (default is 'mindsdb')
                include_tables: List of tables to include for text2sql skills
                ignore_tables: List of tables to ignore for text2sql skills
                include_knowledge_bases: List of knowledge bases to include for text2sql skills
                ignore_knowledge_bases: List of knowledge bases to ignore for text2sql skills

        Returns:
            agent (db.Agents): The created agent

        Raises:
            EntityExistsError: Agent with given name already exists, or skill/model with given name does not exist.
        """
        if project_name is None:
            project_name = default_project
        project = self.project_controller.get(name=project_name)

        agent = self.get_agent(name, project_name)

        if agent is not None:
            raise EntityExistsError("Agent already exists", name)

        # No need to copy params since we're not preserving the original reference
        params = params or {}

        if isinstance(model_name, dict):
            # move into params
            params["model"] = model_name
            model_name = None

        if model_name is not None:
            _, provider = self.check_model_provider(model_name, provider)

        if model_name is None:
            logger.warning("'model_name' param is not provided. Using default global llm model at runtime.")

        # If model_name is not provided, we use default global llm model at runtime
        # Default parameters will be applied at runtime via get_agent_llm_params
        # This allows global default updates to apply to all agents immediately

        # Extract API key if provided in the format <provider>_api_key
        if provider is not None:
            provider_api_key_param = f"{provider.lower()}_api_key"
            if provider_api_key_param in params:
                # Keep the API key in params for the agent to use
                # It will be picked up by get_api_key() in handler_utils.py
                pass

        # Handle generic api_key parameter if provided
        if "api_key" in params:
            # Keep the generic API key in params for the agent to use
            # It will be picked up by get_api_key() in handler_utils.py
            pass

        depreciated_params = [
            "database",
            "knowledge_base_database",
            "include_tables",
            "ignore_tables",
            "include_knowledge_bases",
            "ignore_knowledge_bases",
        ]
        if any(param in params for param in depreciated_params):
            raise ValueError(
                f"Parameters {', '.join(depreciated_params)} are deprecated. "
                "Use 'data' parameter with 'tables' and 'knowledge_bases' keys instead."
            )

        include_tables = None
        include_knowledge_bases = None
        if "data" in params:
            include_knowledge_bases = params["data"].get("knowledge_bases")
            include_tables = params["data"].get("tables")

        # Convert string parameters to lists if needed
        if isinstance(include_tables, str):
            include_tables = [t.strip() for t in include_tables.split(",")]
        if isinstance(include_knowledge_bases, str):
            include_knowledge_bases = [kb.strip() for kb in include_knowledge_bases.split(",")]

        # Auto-create SQL skill if no skills are provided but include_tables or include_knowledge_bases params are provided
        if not skills and (include_tables or include_knowledge_bases):
            skill = self._create_default_sql_skill(
                name,
                project_name,
                include_tables=include_tables,
                include_knowledge_bases=include_knowledge_bases,
            )
            skills = [skill]

        agent = db.Agents(
            name=name,
            project_id=project.id,
            company_id=ctx.company_id,
            user_class=ctx.user_class,
            model_name=model_name,
            provider=provider,
            params=params,
        )

        for skill in skills:
            if isinstance(skill, str):
                skill_name = skill
                parameters = {}
            else:
                parameters = skill.copy()
                skill_name = parameters.pop("name")

            existing_skill = self.skills_controller.get_skill(skill_name, project_name)
            if existing_skill is None:
                db.session.rollback()
                raise ValueError(f"Skill with name does not exist: {skill_name}")

            # Run Data Catalog loader if enabled.
            if include_tables:
                self._run_data_catalog_loader_for_table_entries(include_tables, project_name, skill=existing_skill)
            else:
                self._run_data_catalog_loader_for_skill(existing_skill, project_name, tables=parameters.get("tables"))

            if existing_skill.type == "sql":
                # Add table restrictions if this is a text2sql skill
                if include_tables:
                    parameters["tables"] = include_tables

                # Add knowledge base parameters to both the skill and the association parameters
                if include_knowledge_bases:
                    parameters["include_knowledge_bases"] = include_knowledge_bases
                    if "include_knowledge_bases" not in existing_skill.params:
                        existing_skill.params["include_knowledge_bases"] = include_knowledge_bases
                        flag_modified(existing_skill, "params")

            association = db.AgentSkillsAssociation(parameters=parameters, agent=agent, skill=existing_skill)
            db.session.add(association)

        db.session.add(agent)
        db.session.commit()

        return agent

    def update_agent(
        self,
        agent_name: str,
        project_name: str = default_project,
        name: str = None,
        model_name: Union[str, dict] = None,
        skills_to_add: List[Union[str, dict]] = None,
        skills_to_remove: List[str] = None,
        skills_to_rewrite: List[Union[str, dict]] = None,
        provider: str = None,
        params: Dict[str, str] = None,
    ):
        """
        Updates an agent in the database.

        Parameters:
            agent_name (str): The name of the new agent, or existing agent to update
            project_name (str): The containing project
            name (str): The updated name of the agent
            model_name (str | dict): The name of the existing ML model the agent will use
            skills_to_add (List[Union[str, dict]]): List of skill names to add to the agent, or list of dicts
                 with one of keys is "name", and other is additional parameters for relationship agent<>skill
            skills_to_remove (List[str]): List of skill names to remove from the agent
            skills_to_rewrite (List[Union[str, dict]]): new list of skills for the agent
            provider (str): The provider of the model
            params: (Dict[str, str]): Parameters to use when running the agent

        Returns:
            agent (db.Agents): The created or updated agent

        Raises:
            EntityExistsError: if agent with new name already exists
            EntityNotExistsError: if agent with name or skill not found
            ValueError: if conflict in skills list
        """

        skills_to_add = skills_to_add or []
        skills_to_remove = skills_to_remove or []
        skills_to_rewrite = skills_to_rewrite or []

        if len(skills_to_rewrite) > 0 and (len(skills_to_remove) > 0 or len(skills_to_add) > 0):
            raise ValueError(
                "'skills_to_rewrite' and 'skills_to_add' (or 'skills_to_remove') cannot be used at the same time"
            )

        existing_agent = self.get_agent(agent_name, project_name=project_name)
        if existing_agent is None:
            raise EntityNotExistsError(f"Agent with name not found: {agent_name}")
        existing_params = existing_agent.params or {}

        is_demo = (existing_agent.params or {}).get("is_demo", False)
        if is_demo and (
            (name is not None and name != agent_name)
            or (model_name is not None and existing_agent.model_name != model_name)
            or (provider is not None and existing_agent.provider != provider)
            or (isinstance(params, dict) and len(params) > 0 and "prompt_template" not in params)
        ):
            raise ValueError("It is forbidden to change properties of the demo object")

        if name is not None and name != agent_name:
            if not name.islower():
                raise ValueError(f"The name must be in lower case: {name}")
            # Check to see if updated name already exists
            agent_with_new_name = self.get_agent(name, project_name=project_name)
            if agent_with_new_name is not None:
                raise EntityExistsError(f"Agent with updated name already exists: {name}")
            existing_agent.name = name

        if model_name or provider:
            if isinstance(model_name, dict):
                # move into params
                existing_params["model"] = model_name
                model_name = None

            # check model and provider
            model, provider = self.check_model_provider(model_name, provider)
            # Update model and provider
            existing_agent.model_name = model_name
            existing_agent.provider = provider

        if "data" in params:
            if len(skills_to_add) > 0 or len(skills_to_remove) > 0:
                raise ValueError(
                    "'data' parameter cannot be used with 'skills_to_remove' or 'skills_to_add' parameters"
                )

            include_knowledge_bases = params["data"].get("knowledge_bases")
            include_tables = params["data"].get("tables")

            skill = self._create_default_sql_skill(
                agent_name,
                project_name,
                include_tables=include_tables,
                include_knowledge_bases=include_knowledge_bases,
            )
            skills_to_rewrite = [{"name": skill}]

        # check that all skills exist
        skill_name_to_record_map = {}
        for skill_meta in skills_to_add + skills_to_remove + skills_to_rewrite:
            skill_name = skill_meta["name"] if isinstance(skill_meta, dict) else skill_meta
            if skill_name not in skill_name_to_record_map:
                skill_record = self.skills_controller.get_skill(skill_name, project_name)
                if skill_record is None:
                    raise EntityNotExistsError(f"Skill with name does not exist: {skill_name}")
                skill_name_to_record_map[skill_name] = skill_record

        if len(skills_to_add) > 0 or len(skills_to_remove) > 0:
            skills_to_add = [{"name": x} if isinstance(x, str) else x for x in skills_to_add]
            skills_to_add_names = [x["name"] for x in skills_to_add]

            # there are no intersection between lists
            if not set(skills_to_add_names).isdisjoint(set(skills_to_remove)):
                raise ValueError("Conflict between skills to add and skills to remove.")

            existing_agent_skills_names = [rel.skill.name for rel in existing_agent.skills_relationships]

            # remove skills
            for skill_name in skills_to_remove:
                for rel in existing_agent.skills_relationships:
                    if rel.skill.name == skill_name:
                        db.session.delete(rel)

            # add skills
            for skill_name in set(skills_to_add_names) - set(existing_agent_skills_names):
                # Run Data Catalog loader if enabled for the new skill
                self._run_data_catalog_loader_for_skill(
                    skill_name,
                    project_name,
                    tables=next((x for x in skills_to_add if x["name"] == skill_name), {}).get("tables"),
                )

                skill_parameters = next(x for x in skills_to_add if x["name"] == skill_name).copy()
                del skill_parameters["name"]
                association = db.AgentSkillsAssociation(
                    parameters=skill_parameters, agent=existing_agent, skill=skill_name_to_record_map[skill_name]
                )
                db.session.add(association)

        elif len(skills_to_rewrite) > 0:
            skill_name_to_parameters = {
                x["name"]: {k: v for k, v in x.items() if k != "name"} for x in skills_to_rewrite
            }
            existing_skill_names = set()
            for rel in existing_agent.skills_relationships:
                if rel.skill.name not in skill_name_to_parameters:
                    db.session.delete(rel)
                else:
                    existing_skill_names.add(rel.skill.name)
                    skill_parameters = skill_name_to_parameters[rel.skill.name]

                    # Run Data Catalog loader if enabled for the updated skill
                    self._run_data_catalog_loader_for_skill(
                        rel.skill.name, project_name, tables=skill_parameters.get("tables")
                    )

                    rel.parameters = skill_parameters
                    flag_modified(rel, "parameters")
            for new_skill_name in set(skill_name_to_parameters) - existing_skill_names:
                # Run Data Catalog loader if enabled for the new skill
                self._run_data_catalog_loader_for_skill(
                    new_skill_name,
                    project_name,
                    tables=skill_name_to_parameters[new_skill_name].get("tables"),
                )

                association = db.AgentSkillsAssociation(
                    parameters=skill_name_to_parameters[new_skill_name],
                    agent=existing_agent,
                    skill=skill_name_to_record_map[new_skill_name],
                )
                db.session.add(association)

        if params is not None:
            if params.get("data", {}).get("tables"):
                new_table_entries = set(params["data"]["tables"]) - set(
                    existing_params.get("data", {}).get("tables", [])
                )
                if new_table_entries:
                    # Run Data Catalog loader for new table entries if enabled.
                    self._run_data_catalog_loader_for_table_entries(new_table_entries, project_name)

            # Merge params on update
            existing_params.update(params)
            # Remove None values entirely.
            params = {k: v for k, v in existing_params.items() if v is not None}
            existing_agent.params = params
            # Some versions of SQL Alchemy won't handle JSON updates correctly without this.
            # See: https://docs.sqlalchemy.org/en/20/orm/session_api.html#sqlalchemy.orm.attributes.flag_modified
            flag_modified(existing_agent, "params")
        db.session.commit()

        return existing_agent

    def _run_data_catalog_loader_for_skill(
        self,
        skill: Union[str, db.Skills],
        project_name: str,
        tables: List[str] = None,
    ):
        """
        Runs Data Catalog loader for a skill if enabled in the config.
        This is used to load metadata for SQL skills when they are added or updated.
        """
        if not config.get("data_catalog", {}).get("enabled", False):
            return

        skill = skill if isinstance(skill, db.Skills) else self.skills_controller.get_skill(skill, project_name)
        if skill.type == "sql":
            if "database" in skill.params:
                valid_table_names = skill.params.get("tables") if skill.params.get("tables") else tables
                data_catalog_loader = DataCatalogLoader(
                    database_name=skill.params["database"], table_names=valid_table_names
                )
                data_catalog_loader.load_metadata()
            else:
                raise ValueError(
                    "Data Catalog loading is enabled, but the provided parameters for the new skills are insufficient to load metadata. "
                )

    def _run_data_catalog_loader_for_table_entries(
        self,
        table_entries: List[str],
        project_name: str,
        skill: Union[str, db.Skills] = None,
    ):
        """
        Runs Data Catalog loader for a list of table entries if enabled in the config.
        This is used to load metadata for SQL skills when they are added or updated.
        """
        if not config.get("data_catalog", {}).get("enabled", False):
            return

        skill = skill if isinstance(skill, db.Skills) else self.skills_controller.get_skill(skill, project_name)
        if not skill or skill.type == "sql":
            database_table_map = {}
            for table_entry in table_entries:
                parts = table_entry.split(".", 1)

                # Ensure the table name is in 'database.table' format.
                if len(parts) != 2:
                    logger.warning(
                        f"Invalid table name format: {table_entry}. Expected 'database.table' format."
                        "Metadata will not be loaded for this entry."
                    )
                    continue

                database, table = parts[0], parts[1]

                # Wildcards in database names are not supported at the moment by data catalog loader.
                if "*" in database:
                    logger.warning(
                        f"Invalid database name format: {database}. Wildcards are not supported."
                        "Metadata will not be loaded for this entry."
                    )
                    continue

                # Wildcards in table names are supported either.
                # However, the table name itself can be a wildcard representing all tables.
                if table == "*":
                    table = None
                elif "*" in table:
                    logger.warning(
                        f"Invalid table name format: {table}. Wildcards are not supported."
                        "Metadata will not be loaded for this entry."
                    )
                    continue

                database_table_map[database] = database_table_map.get(database, []) + [table]

            for database_name, table_names in database_table_map.items():
                data_catalog_loader = DataCatalogLoader(database_name=database_name, table_names=table_names)
                data_catalog_loader.load_metadata()

    def delete_agent(self, agent_name: str, project_name: str = default_project):
        """
        Deletes an agent by name.

        Parameters:
            agent_name (str): The name of the agent to delete
            project_name (str): The name of the containing project

        Raises:
            ValueError: Agent does not exist.
        """

        agent = self.get_agent(agent_name, project_name)
        if agent is None:
            raise ValueError(f"Agent with name does not exist: {agent_name}")
        if isinstance(agent.params, dict) and agent.params.get("is_demo") is True:
            raise ValueError("Unable to delete demo object")
        agent.deleted_at = datetime.datetime.now()
        db.session.commit()

    def get_agent_llm_params(self, agent_params: dict):
        """
        Get agent LLM parameters by combining default config with user provided parameters.
        Similar to how knowledge bases handle default parameters.
        """
        combined_model_params = copy.deepcopy(config.get("default_llm", {}))

        if "model" in agent_params:
            model_params = agent_params["model"]
        else:
            # params for LLM can be arbitrary
            model_params = agent_params

        if model_params:
            combined_model_params.update(model_params)

        return combined_model_params

    def get_completion(
        self,
        agent: db.Agents,
        messages: list[Dict[str, str]],
        project_name: str = default_project,
        tools: list[BaseTool] = None,
        stream: bool = False,
        params: dict | None = None,
    ) -> Union[Iterator[object], pd.DataFrame]:
        """
        Queries an agent to get a completion.

        Parameters:
            agent (db.Agents): Existing agent to get completion from
            messages (list[Dict[str, str]]): Chat history to send to the agent
            project_name (str): Project the agent belongs to (default mindsdb)
            tools (list[BaseTool]): Tools to use while getting the completion
            stream (bool): Whether to stream the response
            params (dict | None): params to redefine agent params

        Returns:
            response (Union[Iterator[object], pd.DataFrame]): Completion as a DataFrame or iterator of completion chunks

        Raises:
            ValueError: Agent's model does not exist.
        """
        if stream:
            return self._get_completion_stream(agent, messages, project_name=project_name, tools=tools, params=params)
        from .langchain_agent import LangchainAgent

        model, provider = self.check_model_provider(agent.model_name, agent.provider)
        # update old agents
        if agent.provider is None and provider is not None:
            agent.provider = provider
            db.session.commit()

        # Get agent parameters and combine with default LLM parameters at runtime
        llm_params = self.get_agent_llm_params(agent.params)

        lang_agent = LangchainAgent(agent, model, llm_params=llm_params)
        return lang_agent.get_completion(messages, params=params)

    def _get_completion_stream(
        self,
        agent: db.Agents,
        messages: list[Dict[str, str]],
        project_name: str = default_project,
        tools: list[BaseTool] = None,
        params: dict | None = None,
    ) -> Iterator[object]:
        """
        Queries an agent to get a stream of completion chunks.

        Parameters:
            agent (db.Agents): Existing agent to get completion from
            messages (list[Dict[str, str]]): Chat history to send to the agent
            trace_id (str): ID of Langfuse trace to use
            observation_id (str): ID of parent Langfuse observation to use
            project_name (str): Project the agent belongs to (default mindsdb)
            tools (list[BaseTool]): Tools to use while getting the completion
            params (dict | None): params to redefine agent params

        Returns:
            chunks (Iterator[object]): Completion chunks as an iterator

        Raises:
            ValueError: Agent's model does not exist.
        """
        # For circular dependency.
        from .langchain_agent import LangchainAgent

        model, provider = self.check_model_provider(agent.model_name, agent.provider)

        # update old agents
        if agent.provider is None and provider is not None:
            agent.provider = provider
            db.session.commit()

        # Get agent parameters and combine with default LLM parameters at runtime
        llm_params = self.get_agent_llm_params(agent.params)

        lang_agent = LangchainAgent(agent, model=model, llm_params=llm_params)
        return lang_agent.get_completion(messages, stream=True, params=params)



import json
import shopify
import requests
import pandas as pd
from typing import Text, List, Dict, Any, Set

from mindsdb_sql_parser import ast
from mindsdb.integrations.libs.api_handler import APITable

from mindsdb.integrations.utilities.handlers.query_utilities import SELECTQueryParser, SELECTQueryExecutor
from mindsdb.integrations.utilities.handlers.query_utilities import INSERTQueryParser
from mindsdb.integrations.utilities.handlers.query_utilities import DELETEQueryParser, DELETEQueryExecutor
from mindsdb.integrations.utilities.handlers.query_utilities import UPDATEQueryParser, UPDATEQueryExecutor

from mindsdb.utilities import log

logger = log.getLogger(__name__)


class ProductsTable(APITable):
    """The Shopify Products Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Shopify "GET /products" API endpoint.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Shopify Products matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'products',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        products_df = pd.json_normalize(self.get_products(limit=result_limit))

        select_statement_executor = SELECTQueryExecutor(
            products_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        products_df = select_statement_executor.execute_query()

        return products_df

    def insert(self, query: ast.Insert) -> None:
        """Inserts data into the Shopify "POST /products" API endpoint.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=['title', 'body_html', 'vendor', 'product_type', 'tags', 'status'],
            mandatory_columns=['title'],
            all_mandatory=False
        )
        product_data = insert_statement_parser.parse_query()
        self.create_products(product_data)

    def delete(self, query: ast.Delete) -> None:
        """
        Deletes data from the Shopify "DELETE /products" API endpoint.

        Parameters
        ----------
        query : ast.Delete
           Given SQL DELETE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        delete_statement_parser = DELETEQueryParser(query)
        where_conditions = delete_statement_parser.parse_query()

        products_df = pd.json_normalize(self.get_products())

        delete_query_executor = DELETEQueryExecutor(
            products_df,
            where_conditions
        )

        products_df = delete_query_executor.execute_query()

        product_ids = products_df['id'].tolist()
        self.delete_products(product_ids)

    def update(self, query: ast.Update) -> None:
        """Updates data from the Shopify "PUT /products" API endpoint.

        Parameters
        ----------
        query : ast.Update
           Given SQL UPDATE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        update_statement_parser = UPDATEQueryParser(query)
        values_to_update, where_conditions = update_statement_parser.parse_query()

        products_df = pd.json_normalize(self.get_products())
        update_query_executor = UPDATEQueryExecutor(
            products_df,
            where_conditions
        )

        products_df = update_query_executor.execute_query()
        product_ids = products_df['id'].tolist()
        self.update_products(product_ids, values_to_update)

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_products(limit=1)).columns.tolist()

    def get_products(self, **kwargs) -> List[Dict]:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)
        products = shopify.Product.find(**kwargs)
        return [product.to_dict() for product in products]

    def update_products(self, product_ids: List[int], values_to_update: Dict[Text, Any]) -> None:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)

        for product_id in product_ids:
            product = shopify.Product.find(product_id)
            for key, value in values_to_update.items():
                setattr(product, key, value)
            product.save()
            logger.info(f'Product {product_id} updated')

    def delete_products(self, product_ids: List[int]) -> None:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)

        for product_id in product_ids:
            product = shopify.Product.find(product_id)
            product.destroy()
            logger.info(f'Product {product_id} deleted')

    def create_products(self, product_data: List[Dict[Text, Any]]) -> None:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)

        for product in product_data:
            created_product = shopify.Product.create(product)
            if 'id' not in created_product.to_dict():
                raise Exception('Product creation failed')
            else:
                logger.info(f'Product {created_product.to_dict()["id"]} created')


class CustomersTable(APITable):
    """The Shopify Customers Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Shopify "GET /customers" API endpoint.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Shopify Customers matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        select_statement_parser = SELECTQueryParser(
            query,
            'customers',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        customers_df = pd.json_normalize(self.get_customers(limit=result_limit))

        select_statement_executor = SELECTQueryExecutor(
            customers_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        customers_df = select_statement_executor.execute_query()

        return customers_df

    def insert(self, query: ast.Insert) -> None:
        """Inserts data into the Shopify "POST /customers" API endpoint.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=['first_name', 'last_name', 'email', 'phone', 'tags', 'currency'],
            mandatory_columns=['first_name', 'last_name', 'email', 'phone'],
            all_mandatory=False
        )
        customer_data = insert_statement_parser.parse_query()
        self.create_customers(customer_data)

    def update(self, query: ast.Update) -> None:
        """Updates data in the Shopify "PUT /customers" API endpoint.

        Parameters
        ----------
        query : ast.Update
           Given SQL UPDATE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        update_statement_parser = UPDATEQueryParser(query)
        values_to_update, where_conditions = update_statement_parser.parse_query()

        customers_df = pd.json_normalize(self.get_customers())

        update_query_executor = UPDATEQueryExecutor(
            customers_df,
            where_conditions
        )

        customers_df = update_query_executor.execute_query()

        customer_ids = customers_df['id'].tolist()

        self.update_customers(customer_ids, values_to_update)

    def delete(self, query: ast.Delete) -> None:
        """Deletes data from the Shopify "DELETE /customers" API endpoint.

        Parameters
        ----------
        query : ast.Delete
           Given SQL DELETE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        delete_statement_parser = DELETEQueryParser(query)
        where_conditions = delete_statement_parser.parse_query()

        customers_df = pd.json_normalize(self.get_customers())

        delete_query_executor = DELETEQueryExecutor(
            customers_df,
            where_conditions
        )

        customers_df = delete_query_executor.execute_query()

        customer_ids = customers_df['id'].tolist()
        self.delete_customers(customer_ids)

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_customers(limit=1)).columns.tolist()

    def get_customers(self, **kwargs) -> List[Dict]:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)
        customers = shopify.Customer.find(**kwargs)
        return [customer.to_dict() for customer in customers]

    def create_customers(self, customer_data: List[Dict[Text, Any]]) -> None:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)

        for customer in customer_data:
            created_customer = shopify.Customer.create(customer)
            if 'id' not in created_customer.to_dict():
                raise Exception('Customer creation failed')
            else:
                logger.info(f'Customer {created_customer.to_dict()["id"]} created')

    def update_customers(self, customer_ids: List[int], values_to_update: List[Dict[Text, Any]]) -> None:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)

        for customer_id in customer_ids:
            customer = shopify.Customer.find(customer_id)
            for key, value in values_to_update.items():
                setattr(customer, key, value)
            customer.save()
            logger.info(f'Customer {customer_id} updated')

    def delete_customers(self, customer_ids: List[int]) -> None:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)

        for customer_id in customer_ids:
            customer = shopify.Customer.find(customer_id)
            customer.delete()
            logger.info(f'Customer {customer_id} deleted')


class OrdersTable(APITable):
    """The Shopify Orders Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Shopify "GET /orders" API endpoint.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Shopify Orders matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        select_statement_parser = SELECTQueryParser(
            query,
            'orders',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()
        orders_df = pd.json_normalize(self.get_orders(limit=result_limit))

        select_statement_executor = SELECTQueryExecutor(
            orders_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        orders_df = select_statement_executor.execute_query()

        return orders_df

    def insert(self, query: ast.Insert) -> None:
        """
        Inserts data into the Shopify "POST /orders" API endpoint.

        Parameters
        ----------
        query : ast.Insert
            Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=['address1_ba', 'address2_ba', 'city_ba', 'company_ba', 'country_ba',
                               'country_code_ba', 'first_name_ba', 'last_name_ba', 'latitude_ba',
                               'longitude_ba', 'name_ba', 'phone_ba', 'province_ba', 'province_code_ba',
                               'zip_ba',
                               'address1_sa', 'address2_sa', 'city_sa', 'company_sa',
                               'country_sa', 'country_code_sa', 'first_name_sa', 'last_name_sa',
                               'latitude_sa', 'longitude_sa', 'name_sa', 'phone_sa', 'province_sa',
                               'province_code_sa', 'zip_sa',
                               'amount_dc', 'code_dc', 'type_dc',
                               'gift_card_li', 'grams_li', 'price_li', 'quantity_li', 'title_li',
                               'vendor_li', 'fulfillment_status_li', 'sku_li', 'variant_title_li',
                               'name_li', 'value_li',
                               'price_tl', 'rate_tl', 'title_tl', 'channel_liable_tl',
                               'name_na', 'value_na',
                               'code_sl', 'price_sl', 'discounted_price_sl', 'source_sl',
                               'title_sl',
                               'carrier_identifier_sl', 'requested_fulfillment_service_id_sl',
                               'is_removed_sl',
                               'buyer_accepts_marketing', 'currency', 'email', 'financial_status',
                               'fulfillment_status', 'note', 'phone', 'po_number', 'processed_at',
                               'referring_site', 'source_name', 'source_identifier', 'source_url',
                               'tags', 'taxes_included', 'test', 'total_tax', 'total_weight'],
            mandatory_columns=['price_li', 'title_li'],
            all_mandatory=False
        )
        order_data = insert_statement_parser.parse_query()
        self.create_orders(order_data)

    def update(self, query: ast.Update) -> None:
        """Updates data in the Shopify "PUT /orders" API endpoint.

        Parameters
        ----------
        query : ast.Update
           Given SQL UPDATE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        update_statement_parser = UPDATEQueryParser(query)
        values_to_update, where_conditions = update_statement_parser.parse_query()
        orders_df = pd.json_normalize(self.get_orders())

        update_statement_executor = UPDATEQueryExecutor(
            orders_df,
            where_conditions
        )
        orders_df = update_statement_executor.execute_query()
        orders_ids = orders_df['id'].tolist()
        self.update_orders(orders_ids, values_to_update)

    def delete(self, query: ast.Delete) -> None:
        """Deletes data from the Shopify "DELETE /orders" API endpoint.

        Parameters
        ----------
        query : ast.Delete
           Given SQL DELETE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        delete_statement_parser = DELETEQueryParser(query)
        where_conditions = delete_statement_parser.parse_query()

        orders_df = pd.json_normalize(self.get_orders())

        delete_query_executor = DELETEQueryExecutor(
            orders_df,
            where_conditions
        )

        orders_df = delete_query_executor.execute_query()

        order_ids = orders_df['id'].tolist()
        self.delete_orders(order_ids)

    def update_orders(self, order_ids: List[int], values_to_update: Dict[Text, Any]) -> None:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)

        for order_id in order_ids:
            order = shopify.Order.find(order_id)
            for key, value in values_to_update.items():
                setattr(order, key, value)
            order.save()
            logger.info(f'Order {order_id} updated')

    def create_orders(self, order_data: List[Dict[Text, Any]]) -> None:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)
        # separate columns by API object
        line_items_columns = {'gift_card_li', 'grams_li', 'price_li', 'quantity_li', 'title_li',
                              'vendor_li', 'fulfillment_status_li', 'sku_li', 'variant_title_li'}
        billing_address_columns = {'address1_ba', 'address2_ba', 'city_ba', 'company_ba',
                                   'country_ba', 'country_code_ba', 'first_name_ba', 'last_name_ba',
                                   'latitude_ba', 'longitude_ba', 'name_ba', 'phone_ba',
                                   'province_ba', 'province_code_ba', 'zip_ba'}
        shipping_address_columns = {'address1_sa', 'address2_sa', 'city_sa', 'company_sa',
                                    'country_sa', 'country_code_sa', 'first_name_sa', 'last_name_sa',
                                    'latitude_sa', 'longitude_sa', 'name_sa', 'phone_sa',
                                    'province_sa', 'province_code_sa', 'zip_sa'}
        discount_codes_columns = {'amount_dc', 'code_dc', 'type_dc'}
        tax_lines_columns = {'price_tl', 'rate_tl', 'title_tl', 'channel_liable_tl'}
        note_attributes_columns = {'name_na', 'value_na'}
        shipping_lines_columns = {'code_sl', 'price_sl', 'discounted_price_sl',
                                  'source_sl', 'title_sl', 'carrier_identifier_sl',
                                  'requested_fulfillment_service_id_sl', 'is_removed_sl'}
        line_items_properties_columns = {'name_li', 'value_li'}
        all_columns = (line_items_columns | billing_address_columns | shipping_address_columns
                       | discount_codes_columns | tax_lines_columns | note_attributes_columns
                       | shipping_lines_columns | line_items_properties_columns)
        modified_order_data = []

        for order in order_data:
            # separate values by object
            order_data_trimmed = {key: val for key, val in order.items()
                                  if key not in all_columns}
            line_items_data = OrdersTable._extract_data_helper(order, line_items_columns)
            billing_address_data = OrdersTable._extract_data_helper(order, billing_address_columns)
            shipping_address_data = OrdersTable._extract_data_helper(order, shipping_address_columns)
            discount_codes_data = OrdersTable._extract_data_helper(order, discount_codes_columns)
            tax_lines_data = OrdersTable._extract_data_helper(order, tax_lines_columns)
            note_attributes_data = OrdersTable._extract_data_helper(order, note_attributes_columns)
            shipping_lines_data = OrdersTable._extract_data_helper(order, shipping_lines_columns)
            line_items_properties_data = OrdersTable._extract_data_helper(order, line_items_properties_columns)

            # add sub-arrays to line items object
            line_items_data['properties'] = [line_items_properties_data]

            # add JSON and array objects to dictionary
            order_data_trimmed['billing_address'] = json.loads(json.dumps(billing_address_data))
            order_data_trimmed['shipping_address'] = json.loads(json.dumps(shipping_address_data))
            order_data_trimmed['line_items'] = json.loads(json.dumps([line_items_data]))
            order_data_trimmed['discount_codes'] = json.loads(json.dumps([discount_codes_data]))
            order_data_trimmed['tax_lines'] = json.loads(json.dumps([tax_lines_data]))
            order_data_trimmed['note_attributes'] = json.loads(json.dumps([note_attributes_data]))
            order_data_trimmed['shipping_lines'] = json.loads(json.dumps([shipping_lines_data]))

            modified_order_data.append(order_data_trimmed)

        for order in modified_order_data:

            created_order = shopify.Order.create(order)
            if 'id' not in created_order.to_dict():
                raise Exception('Order creation failed')

            logger.info(f'Order {created_order.to_dict()["id"]} created')

    @staticmethod
    def _extract_data_helper(order: Dict, columns: Set, subscript_len: int = 3) -> Dict:
        strip_index = subscript_len * -1
        return {key[:strip_index]: val for key, val in order.items() if key in columns}

    def delete_orders(self, order_ids: List[int]) -> None:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)

        for order_id in order_ids:
            order = shopify.Order.find(order_id)
            order.destroy()
            logger.info(f'Order {order_id} deleted')

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_orders(limit=1)).columns.tolist()

    def get_orders(self, **kwargs) -> List[Dict]:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)
        orders = shopify.Order.find(**kwargs)
        return [order.to_dict() for order in orders]


class InventoryLevelTable(APITable):
    """The Shopify Inventory Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Shopify "GET /inventory" API endpoint.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame

            Shopify Inventory matching the inventory_item_id or/and location_id

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        select_statement_parser = SELECTQueryParser(
            query,
            'inventory_level',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'inventory_item_ids':
                if op == '=':
                    search_params["inventory_item_ids"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for inventory_item_ids column.")
            elif arg1 == 'location_ids':
                if op == '=':
                    search_params["location_ids"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for location_ids column.")
            elif arg1 in ['available', 'updated_at']:
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("inventory_item_ids" in search_params) or ("location_ids" in search_params)

        if not filter_flag:
            raise NotImplementedError("inventory_item_ids or location_ids column has to be present in where clause.")

        search_params["limit"] = result_limit

        inventory_df = pd.json_normalize(self.get_inventory(search_params))

        self.clean_selected_columns(selected_columns)

        select_statement_executor = SELECTQueryExecutor(
            inventory_df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions
        )
        inventory_df = select_statement_executor.execute_query()

        return inventory_df

    def clean_selected_columns(self, selected_cols) -> List[Text]:
        if "inventory_item_ids" in selected_cols:
            selected_cols.remove("inventory_item_ids")
            selected_cols.append("inventory_item_id")
        if "location_ids" in selected_cols:
            selected_cols.remove("location_ids")
            selected_cols.append("location_id")

    def get_columns(self) -> List[Text]:
        return ["inventory_item_ids", "location_ids", "available", "updated_at"]

    def get_inventory(self, kwargs) -> List[Dict]:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)
        inventories = shopify.InventoryLevel.find(**kwargs)
        return [inventory.to_dict() for inventory in inventories]


class LocationTable(APITable):
    """The Shopify Location Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Shopify "GET /locations" API endpoint.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame

            Shopify Locations matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        select_statement_parser = SELECTQueryParser(
            query,
            'locations',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        locations_df = pd.json_normalize(self.get_locations(limit=result_limit))

        select_statement_executor = SELECTQueryExecutor(
            locations_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )

        locations_df = select_statement_executor.execute_query()

        return locations_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_locations(limit=1)).columns.tolist()

    def get_locations(self, **kwargs) -> List[Dict]:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)
        locations = shopify.Location.find(**kwargs)
        return [location.to_dict() for location in locations]


class CustomerReviews(APITable):
    """The Shopify Customer Reviews Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Yotpo "GET https://api.yotpo.com/v1/apps/{app_key}/reviews?utoken={utoken}" API endpoint.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame

            Shopify customer reviews matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        select_statement_parser = SELECTQueryParser(
            query,
            'customer_reviews',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        customer_reviews_df = pd.json_normalize(self.get_customer_reviews(limit=result_limit))

        select_statement_executor = SELECTQueryExecutor(
            customer_reviews_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )

        customer_reviews_df = select_statement_executor.execute_query()

        return customer_reviews_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_customer_reviews(limit=1)).columns.tolist()

    def get_customer_reviews(self, **kwargs) -> List[Dict]:
        if self.handler.yotpo_app_key is None or self.handler.yotpo_access_token is None:
            raise Exception("You need to provide 'yotpo_app_key' and 'yotpo_access_token' to retrieve customer reviews.")
        url = f"https://api.yotpo.com/v1/apps/{self.handler.yotpo_app_key}/reviews?count=0&utoken={self.handler.yotpo_access_token}"
        headers = {
            "accept": "application/json",
            "Content-Type": "application/json"
        }
        json_response = requests.get(url, headers=headers).json()
        return [review for review in json_response['reviews']] if 'reviews' in json_response else []


class CarrierServiceTable(APITable):
    """The Shopify carrier service Table implementation. Example carrier services like usps, dhl etc."""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Shopify "GET /carrier_services" API endpoint.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame

            Shopify Carrier service info

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        select_statement_parser = SELECTQueryParser(
            query,
            'carrier_service',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        carrier_service_df = pd.json_normalize(self.get_carrier_service())

        select_statement_executor = SELECTQueryExecutor(
            carrier_service_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )

        carrier_service_df = select_statement_executor.execute_query()

        return carrier_service_df

    def insert(self, query: ast.Insert) -> None:
        """Inserts data into the Shopify "POST /carrier_services" API endpoint.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=['name', 'callback_url', 'service_discovery'],
            mandatory_columns=['name', 'callback_url', 'service_discovery'],
            all_mandatory=True
        )
        carrier_service_data = insert_statement_parser.parse_query()
        self.create_carrier_service(carrier_service_data)

    def delete(self, query: ast.Delete) -> None:
        """
        Deletes data from the Shopify "DELETE /carrier_services" API endpoint.

        Parameters
        ----------
        query : ast.Delete
           Given SQL DELETE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        delete_statement_parser = DELETEQueryParser(query)
        where_conditions = delete_statement_parser.parse_query()

        carrier_services_df = pd.json_normalize(self.get_carrier_service())

        delete_query_executor = DELETEQueryExecutor(
            carrier_services_df,
            where_conditions
        )

        carrier_services_df = delete_query_executor.execute_query()

        carrier_service_ids = carrier_services_df['id'].tolist()
        self.delete_carrier_services(carrier_service_ids)

    def update(self, query: ast.Update) -> None:
        """Updates data from the Shopify "PUT /carrier_services" API endpoint.

        Parameters
        ----------
        query : ast.Update
           Given SQL UPDATE query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        update_statement_parser = UPDATEQueryParser(query)
        values_to_update, where_conditions = update_statement_parser.parse_query()
        carrier_services_df = pd.json_normalize(self.get_carrier_service())
        update_query_executor = UPDATEQueryExecutor(
            carrier_services_df,
            where_conditions
        )

        carrier_services_df = update_query_executor.execute_query()
        carrier_service_ids = carrier_services_df['id'].tolist()
        self.update_carrier_service(carrier_service_ids, values_to_update)

    def get_columns(self) -> List[Text]:
        return ["id", "name", "active", "service_discovery", "carrier_service_type", "admin_graphql_api_id"]

    def get_carrier_service(self) -> List[Dict]:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)
        services = shopify.CarrierService.find()
        return [service.to_dict() for service in services]

    def create_carrier_service(self, carrier_service_data: List[Dict[Text, Any]]) -> None:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)

        for carrier_service in carrier_service_data:
            created_carrier_service = shopify.CarrierService.create(carrier_service)
            if 'id' not in created_carrier_service.to_dict():
                raise Exception('Product creation failed')
            else:
                logger.info(f'Product {created_carrier_service.to_dict()["id"]} created')

    def delete_carrier_services(self, carrier_service_ids: List[int]) -> None:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)

        for carrier_service_id in carrier_service_ids:
            product = shopify.CarrierService.find(carrier_service_id)
            product.destroy()
            logger.info(f'Carrier Service {carrier_service_id} deleted')

    def update_carrier_service(self, carrier_service_id: int, values_to_update: Dict[Text, Any]) -> None:
        # Update the carrier service using the Shopify API
        session = self.handler.connect()
        shopify.ShopifyResource.activate_session(session)

        carrier_service = shopify.CarrierService.find(carrier_service_id)
        for key, value in values_to_update.items():
            setattr(carrier_service, key, value)
        carrier_service.save()
        logger.info(f'Carrier Service {carrier_service_id} updated')


class ShippingZoneTable(APITable):
    """The Shopify shipping zone Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Shopify "GET /shipping_zone" API endpoint.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame

            Shopify Shipping Zone info

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        select_statement_parser = SELECTQueryParser(
            query,
            'shipping_zone',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        shipping_zone_df = pd.json_normalize(self.get_shipping_zone(), record_path="countries", meta=["id", "name"], record_prefix="countries_")

        select_statement_executor = SELECTQueryExecutor(
            shipping_zone_df,
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit
        )

        shipping_zone_df = select_statement_executor.execute_query()

        return shipping_zone_df

    def get_columns(self) -> List[Text]:
        return ['countries_id', 'countries_name', 'countries_tax', 'countries_code', 'countries_tax_name', 'countries_shipping_zone_id', 'countries_provinces', 'id', 'name']

    def clean_response(self, res):
        temp = {}
        temp["id"] = res["id"]
        temp["name"] = res["name"]
        temp["countries"] = res["countries"]
        return temp

    def get_shipping_zone(self) -> List[Dict]:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)
        zones = shopify.ShippingZone.find()
        return [self.clean_response(zone.to_dict()) for zone in zones]


class SalesChannelTable(APITable):
    """The Shopify Sales Channel Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls data from the Shopify "GET /publication API endpoint, as Channel API endpoint is deprecated


        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Shopify Products matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'sales_channel',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        sales_channel_df = pd.json_normalize(self.get_sales_channel(limit=result_limit))

        select_statement_executor = SELECTQueryExecutor(
            sales_channel_df,
            selected_columns,
            where_conditions,
            order_by_conditions,
            result_limit
        )
        sales_channel_df = select_statement_executor.execute_query()
        return sales_channel_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_sales_channel(limit=1)).columns.tolist()

    def get_sales_channel(self, **kwargs) -> List[Dict]:
        api_session = self.handler.connect()
        shopify.ShopifyResource.activate_session(api_session)
        sales_channels = shopify.Publication.find(**kwargs)
        return [sales_channel.to_dict() for sales_channel in sales_channels]



from dataclasses import dataclass, astuple
from datetime import datetime


class TABLES_ROW_TYPE:
    __slots__ = ()
    BASE_TABLE = 'BASE TABLE'
    VIEW = 'VIEW'
    SYSTEM_VIEW = 'SYSTEM VIEW'


TABLES_ROW_TYPE = TABLES_ROW_TYPE()


@dataclass(slots=True)
class TablesRow:
    TABLE_CATALOG: str = 'def'
    TABLE_SCHEMA: str = 'information_schema'
    TABLE_NAME: str = None
    TABLE_TYPE: str = TABLES_ROW_TYPE.BASE_TABLE
    ENGINE: str = None
    VERSION: int = None
    ROW_FORMAT: str = None
    TABLE_ROWS: int = 0
    AVG_ROW_LENGTH: int = 0
    DATA_LENGTH: int = 0
    MAX_DATA_LENGTH: int = 0
    INDEX_LENGTH: int = 0
    DATA_FREE: int = 0
    AUTO_INCREMENT: int = None
    CREATE_TIME: datetime = datetime(2024, 1, 1)
    UPDATE_TIME: datetime = datetime(2024, 1, 1)
    CHECK_TIME: datetime = datetime(2024, 1, 1)
    TABLE_COLLATION: str = None
    CHECKSUM: int = None
    CREATE_OPTIONS: str = None
    TABLE_COMMENT: str = ''

    def to_list(self) -> list:
        return list(astuple(self))

    @staticmethod
    def from_dict(data: dict):
        del_keys = []
        data = {k.upper(): v for k, v in data.items()}

        # table is different column
        if 'TABLE_NAME' not in data and 'NAME' in data:
            data['TABLE_NAME'] = data['NAME']

        for key in data:
            if key not in TablesRow.__dataclass_fields__:
                del_keys.append(key)

        for key in del_keys:
            del data[key]

        return TablesRow(**data)



import os
import mimetypes
import threading
import traceback
import webbrowser

from pathlib import Path
from http import HTTPStatus


import requests
from flask import Flask, url_for, make_response, request, send_from_directory
from flask.json import dumps
from flask_compress import Compress
from flask_restx import Api
from werkzeug.exceptions import HTTPException
from packaging.version import Version, parse as parse_version

from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.api.http.gui import update_static
from mindsdb.api.http.utils import http_error
from mindsdb.api.http.namespaces.agents import ns_conf as agents_ns
from mindsdb.api.http.namespaces.analysis import ns_conf as analysis_ns
from mindsdb.api.http.namespaces.auth import ns_conf as auth_ns
from mindsdb.api.http.namespaces.chatbots import ns_conf as chatbots_ns
from mindsdb.api.http.namespaces.jobs import ns_conf as jobs_ns
from mindsdb.api.http.namespaces.config import ns_conf as conf_ns
from mindsdb.api.http.namespaces.databases import ns_conf as databases_ns
from mindsdb.api.http.namespaces.default import ns_conf as default_ns
from mindsdb.api.http.namespaces.file import ns_conf as file_ns
from mindsdb.api.http.namespaces.handlers import ns_conf as handlers_ns
from mindsdb.api.http.namespaces.knowledge_bases import ns_conf as knowledge_bases_ns
from mindsdb.api.http.namespaces.models import ns_conf as models_ns
from mindsdb.api.http.namespaces.projects import ns_conf as projects_ns
from mindsdb.api.http.namespaces.skills import ns_conf as skills_ns
from mindsdb.api.http.namespaces.sql import ns_conf as sql_ns
from mindsdb.api.http.namespaces.tab import ns_conf as tab_ns
from mindsdb.api.http.namespaces.tree import ns_conf as tree_ns
from mindsdb.api.http.namespaces.views import ns_conf as views_ns
from mindsdb.api.http.namespaces.util import ns_conf as utils_ns
from mindsdb.api.http.namespaces.webhooks import ns_conf as webhooks_ns
from mindsdb.interfaces.database.integrations import integration_controller
from mindsdb.interfaces.database.database import DatabaseController
from mindsdb.interfaces.file.file_controller import FileController
from mindsdb.interfaces.jobs.jobs_controller import JobsController
from mindsdb.interfaces.storage import db
from mindsdb.metrics.server import init_metrics
from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.json_encoder import CustomJSONProvider
from mindsdb.utilities.ps import is_pid_listen_port, wait_func_is_true
from mindsdb.utilities.sentry import sentry_sdk  # noqa: F401
from mindsdb.utilities.otel import trace  # noqa: F401
from mindsdb.api.common.middleware import verify_pat

logger = log.getLogger(__name__)


class _NoOpFlaskInstrumentor:
    def instrument_app(self, app):
        pass


class _NoOpRequestsInstrumentor:
    def instrument(self):
        pass


try:
    from opentelemetry.instrumentation.flask import FlaskInstrumentor
    from opentelemetry.instrumentation.requests import RequestsInstrumentor
except ImportError:
    logger.debug(
        "OpenTelemetry is not avaiable. Please run `pip install -r requirements/requirements-opentelemetry.txt` to use it."
    )
    FlaskInstrumentor = _NoOpFlaskInstrumentor
    RequestsInstrumentor = _NoOpRequestsInstrumentor


class Swagger_Api(Api):
    """
    This is a modification of the base Flask Restplus Api class due to the issue described here
    https://github.com/noirbizarre/flask-restplus/issues/223
    """

    @property
    def specs_url(self):
        return url_for(self.endpoint("specs"), _external=False)


def custom_output_json(data, code, headers=None):
    resp = make_response(dumps(data, cls=CustomJSONProvider), code)
    resp.headers.extend(headers or {})
    return resp


def get_last_compatible_gui_version() -> Version:
    logger.debug("Getting last compatible frontend..")
    try:
        res = requests.get(
            "https://mindsdb-web-builds.s3.amazonaws.com/compatible-config.json",
            timeout=5,
        )
    except (ConnectionError, requests.exceptions.ConnectionError) as e:
        logger.error(f"Is no connection. {e}")
        return False
    except Exception as e:
        logger.error(f"Is something wrong with getting compatible-config.json: {e}")
        return False

    if res.status_code != 200:
        logger.error(f"Cant get compatible-config.json: returned status code = {res.status_code}")
        return False

    try:
        versions = res.json()
    except Exception as e:
        logger.error(f"Cant decode compatible-config.json: {e}")
        return False

    current_mindsdb_lv = parse_version(mindsdb_version)

    try:
        gui_versions = {}
        max_mindsdb_lv = None
        max_gui_lv = None
        for el in versions["mindsdb"]:
            if el["mindsdb_version"] is None:
                gui_lv = parse_version(el["gui_version"])
            else:
                mindsdb_lv = parse_version(el["mindsdb_version"])
                gui_lv = parse_version(el["gui_version"])
                if mindsdb_lv.base_version not in gui_versions or gui_lv > gui_versions[mindsdb_lv.base_version]:
                    gui_versions[mindsdb_lv.base_version] = gui_lv
                if max_mindsdb_lv is None or max_mindsdb_lv < mindsdb_lv:
                    max_mindsdb_lv = mindsdb_lv
            if max_gui_lv is None or max_gui_lv < gui_lv:
                max_gui_lv = gui_lv

        all_mindsdb_lv = [parse_version(x) for x in gui_versions.keys()]
        all_mindsdb_lv.sort()

        if current_mindsdb_lv.base_version in gui_versions:
            gui_version_lv = gui_versions[current_mindsdb_lv.base_version]
        elif current_mindsdb_lv > all_mindsdb_lv[-1]:
            gui_version_lv = max_gui_lv
        else:
            lower_versions = {
                key: value for key, value in gui_versions.items() if parse_version(key) < current_mindsdb_lv
            }
            if len(lower_versions) == 0:
                gui_version_lv = gui_versions[all_mindsdb_lv[0].base_version]
            else:
                all_lower_versions = [parse_version(x) for x in lower_versions.keys()]
                gui_version_lv = gui_versions[all_lower_versions[-1].base_version]
    except Exception as e:
        logger.error(f"Error in compatible-config.json structure: {e}")
        return False

    logger.debug(f"Last compatible frontend version: {gui_version_lv}.")
    return gui_version_lv


def get_current_gui_version() -> Version:
    logger.debug("Getting current frontend version...")
    config = Config()
    static_path = Path(config["paths"]["static"])
    version_txt_path = static_path.joinpath("version.txt")

    current_gui_version = None
    if version_txt_path.is_file():
        with open(version_txt_path, "rt") as f:
            current_gui_version = f.readline()

    current_gui_lv = None if current_gui_version is None else parse_version(current_gui_version)
    logger.debug(f"Current frontend version: {current_gui_lv}.")

    return current_gui_lv


def initialize_static():
    logger.debug("Initializing static..")
    config = Config()
    last_gui_version_lv = get_last_compatible_gui_version()
    current_gui_version_lv = get_current_gui_version()
    required_gui_version = config["gui"].get("version")

    if required_gui_version is not None:
        required_gui_version_lv = parse_version(required_gui_version)
        success = True
        if current_gui_version_lv is None or required_gui_version_lv != current_gui_version_lv:
            logger.debug("Updating gui..")
            success = update_static(required_gui_version_lv)
    else:
        if last_gui_version_lv is False:
            return False

        # ignore versions like '23.9.2.2'
        if current_gui_version_lv is not None and len(current_gui_version_lv.release) < 3:
            if current_gui_version_lv == last_gui_version_lv:
                return True
        logger.debug("Updating gui..")
        success = update_static(last_gui_version_lv)

    if db.session:
        db.session.close()
    return success


def initialize_app(config, no_studio):
    static_root = config["paths"]["static"]
    logger.debug(f"Static route: {static_root}")
    gui_exists = Path(static_root).joinpath("index.html").is_file()
    logger.debug(f"Does GUI already exist.. {'YES' if gui_exists else 'NO'}")
    init_static_thread = None
    if no_studio is False and (config["gui"]["autoupdate"] is True or gui_exists is False):
        init_static_thread = threading.Thread(target=initialize_static, name="initialize_static")
        init_static_thread.start()

    # Wait for static initialization.
    if not no_studio and init_static_thread is not None:
        init_static_thread.join()

    app, api = initialize_flask(config, init_static_thread, no_studio)
    Compress(app)

    initialize_interfaces(app)

    if os.path.isabs(static_root) is False:
        static_root = os.path.join(os.getcwd(), static_root)
    static_root = Path(static_root)

    @app.route("/", defaults={"path": ""}, methods=["GET"])
    @app.route("/<path:path>", methods=["GET"])
    def root_index(path):
        if path.startswith("api/"):
            return http_error(
                HTTPStatus.NOT_FOUND,
                "Not found",
                "The endpoint you are trying to access does not exist on the server.",
            )

        # Normalize the path.
        full_path = os.path.normpath(os.path.join(static_root, path))

        # Check for directory traversal attacks.
        if not full_path.startswith(str(static_root)):
            return http_error(
                HTTPStatus.FORBIDDEN,
                "Forbidden",
                "You are not allowed to access the requested resource.",
            )

        if os.path.isfile(full_path):
            return send_from_directory(static_root, path)
        else:
            return send_from_directory(static_root, "index.html")

    protected_namespaces = [
        tab_ns,
        utils_ns,
        conf_ns,
        file_ns,
        sql_ns,
        analysis_ns,
        handlers_ns,
        tree_ns,
        projects_ns,
        databases_ns,
        views_ns,
        models_ns,
        chatbots_ns,
        skills_ns,
        agents_ns,
        jobs_ns,
        knowledge_bases_ns,
    ]

    for ns in protected_namespaces:
        api.add_namespace(ns)
    api.add_namespace(default_ns)
    api.add_namespace(auth_ns)
    api.add_namespace(webhooks_ns)

    @api.errorhandler(Exception)
    def handle_exception(e):
        logger.error(f"http exception: {e}")
        # pass through HTTP errors
        # NOTE flask_restx require 'message', also it modyfies 'application/problem+json' to 'application/json'
        if isinstance(e, HTTPException):
            return (
                {"title": e.name, "detail": e.description, "message": e.description},
                e.code,
                {"Content-Type": "application/problem+json"},
            )
        return (
            {
                "title": getattr(type(e), "__name__") or "Unknown error",
                "detail": str(e),
                "message": str(e),
            },
            500,
            {"Content-Type": "application/problem+json"},
        )

    @app.teardown_appcontext
    def remove_session(*args, **kwargs):
        db.session.remove()

    @app.before_request
    def before_request():
        logger.debug(f"HTTP {request.method}: {request.path}")
        ctx.set_default()
        config = Config()

        h = request.headers.get("Authorization")
        if not h or not h.startswith("Bearer "):
            bearer = None
        else:
            bearer = h.split(" ", 1)[1].strip() or None

        # region routes where auth is required
        if (
            config["auth"]["http_auth_enabled"] is True
            and any(request.path.startswith(f"/api{ns.path}") for ns in protected_namespaces)
            and verify_pat(bearer) is False
        ):
            logger.debug(f"Auth failed for path {request.path}")
            return http_error(
                HTTPStatus.UNAUTHORIZED,
                "Unauthorized",
                "Authorization is required to complete the request",
            )
        # endregion

        company_id = request.headers.get("company-id")
        user_class = request.headers.get("user-class")

        try:
            email_confirmed = int(request.headers.get("email-confirmed", 1))
        except Exception:
            email_confirmed = 1

        try:
            user_id = int(request.headers.get("user-id", 0))
        except Exception:
            user_id = 0

        if company_id is not None:
            try:
                company_id = int(company_id)
            except Exception as e:
                logger.error(f"Could not parse company id: {company_id} | exception: {e}")
                company_id = None

        if user_class is not None:
            try:
                user_class = int(user_class)
            except Exception as e:
                logger.error(f"Could not parse user_class: {user_class} | exception: {e}")
                user_class = 0
        else:
            user_class = 0

        ctx.user_id = user_id
        ctx.company_id = company_id
        ctx.user_class = user_class
        ctx.email_confirmed = email_confirmed

    logger.debug("Done initializing app.")
    return app


def initialize_flask(config, init_static_thread, no_studio):
    logger.debug("Initializing flask..")
    # region required for windows https://github.com/mindsdb/mindsdb/issues/2526
    mimetypes.add_type("text/css", ".css")
    mimetypes.add_type("text/javascript", ".js")
    # endregion

    kwargs = {}
    if no_studio is not True:
        static_path = os.path.join(config["paths"]["static"], "static/")
        if os.path.isabs(static_path) is False:
            static_path = os.path.join(os.getcwd(), static_path)
        kwargs["static_url_path"] = "/static"
        kwargs["static_folder"] = static_path
        logger.debug(f"Static path: {static_path}")

    app = Flask(__name__, **kwargs)
    init_metrics(app)

    # Instrument Flask app and requests using either real or no-op instrumentors
    FlaskInstrumentor().instrument_app(app)
    RequestsInstrumentor().instrument()

    app.config["SEND_FILE_MAX_AGE_DEFAULT"] = 60
    app.config["SWAGGER_HOST"] = "http://localhost:8000/mindsdb"
    app.json = CustomJSONProvider()

    authorizations = {"apikey": {"type": "apiKey", "in": "header", "name": "Authorization"}}

    logger.debug("Creating swagger API..")
    api = Swagger_Api(
        app,
        authorizations=authorizations,
        security=["apikey"],
        url_prefix=":8000",
        prefix="/api",
        doc="/doc/",
    )

    api.representations["application/json"] = custom_output_json

    port = config["api"]["http"]["port"]
    host = config["api"]["http"]["host"]

    # NOTE rewrite it, that hotfix to see GUI link
    if not no_studio:
        if host in ("", "0.0.0.0"):
            url = f"http://127.0.0.1:{port}/"
        else:
            url = f"http://{host}:{port}/"
        logger.info(f" - GUI available at {url}")

        pid = os.getpid()
        thread = threading.Thread(
            target=_open_webbrowser,
            args=(url, pid, port, init_static_thread, config["paths"]["static"]),
            daemon=True,
            name="open_webbrowser",
        )
        thread.start()

    return app, api


def initialize_interfaces(app):
    app.integration_controller = integration_controller
    app.database_controller = DatabaseController()
    app.file_controller = FileController()
    app.jobs_controller = JobsController()
    config = Config()
    app.config_obj = config


def _open_webbrowser(url: str, pid: int, port: int, init_static_thread, static_folder):
    """Open webbrowser with url when http service is started.

    If some error then do nothing.
    """
    if init_static_thread is not None:
        init_static_thread.join()
    try:
        is_http_active = wait_func_is_true(func=is_pid_listen_port, timeout=15, pid=pid, port=port)
        if is_http_active:
            webbrowser.open(url)
    except Exception as e:
        logger.error(f"Failed to open {url} in webbrowser with exception {e}")
        logger.error(traceback.format_exc())
    db.session.close()



from http import HTTPStatus

from flask import request
from flask_restx import Resource

from mindsdb.api.http.namespaces.configs.projects import ns_conf
from mindsdb.interfaces.agents.agents_controller import AgentsController
from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.api.http.utils import http_error
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.interfaces.chatbot.chatbot_controller import ChatBotController
from mindsdb.interfaces.model.functions import PredictorRecordNotFound
from mindsdb.interfaces.storage.db import Predictor
from mindsdb.utilities.exception import EntityNotExistsError


def create_chatbot(project_name, name, chatbot):
    if name is None:
        return http_error(HTTPStatus.BAD_REQUEST, "Missing field", 'Missing "name" field for chatbot')

    if not name.islower():
        return http_error(HTTPStatus.BAD_REQUEST, "Wrong name", f"The name must be in lower case: {name}")

    model_name = chatbot.get("model_name", None)
    agent_name = chatbot.get("agent_name", None)
    if model_name is None and agent_name is None:
        return http_error(
            HTTPStatus.BAD_REQUEST,
            "Missing field",
            'Must include either "model_name" or "agent_name" field for chatbot',
        )

    session_controller = SessionController()

    if "database_id" in chatbot or "database_name" in chatbot or ("db_engine" in chatbot and "db_params" in chatbot):
        try:
            database_id = get_or_create_database_for_chatbot(chatbot, session_controller)
        except ValueError as value_error:
            return http_error(HTTPStatus.NOT_FOUND, "Database not found", str(value_error))

    else:
        return http_error(
            HTTPStatus.BAD_REQUEST,
            "Missing field",
            'Missing "database_id" or ("db_engine" and "database_param") fields for chatbot',
        )

    is_running = chatbot.get("is_running", False)
    params = chatbot.get("params", {})

    chatbot_controller = ChatBotController()

    # Chatbot can't already exist.
    # TODO all checks should be inside of controller

    try:
        existing_chatbot = chatbot_controller.get_chatbot(name, project_name=project_name)
    except EntityNotExistsError:
        # Project must exist.
        return http_error(HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist")
    if existing_chatbot is not None:
        return http_error(
            HTTPStatus.CONFLICT,
            "Chatbot already exists",
            f"Chatbot with name {name} already exists. Please use a different name",
        )

    # Model and agent need to exist.
    if agent_name is not None:
        agent = AgentsController().get_agent(agent_name, project_name)
        if agent is None:
            return http_error(HTTPStatus.NOT_FOUND, "Agent not found", f"Agent with name {agent_name} not found")
        model_name = agent.model_name

    model_name_no_version, version = Predictor.get_name_and_version(model_name)
    try:
        session_controller.model_controller.get_model(model_name_no_version, version=version, project_name=project_name)
    except PredictorRecordNotFound:
        return http_error(HTTPStatus.NOT_FOUND, "Model not found", f"Model with name {model_name} not found")

    created_chatbot = chatbot_controller.add_chatbot(
        name,
        project_name,
        model_name=model_name,
        agent_name=agent_name,
        database_id=database_id,
        is_running=is_running,
        params=params,
    )
    return created_chatbot.as_dict(), HTTPStatus.CREATED


def get_or_create_database_for_chatbot(chatbot: dict, session_controller: SessionController) -> int:
    """
    Get or create a database for a chatbot, based on the chatbot configuration provided in the request.

    Args:
        chatbot (dict): The chatbot configuration.
        session_controller (SessionController): The session controller.

    Returns:
        int: The database ID.
    """
    if "database_id" in chatbot:
        database_record = session_controller.integration_controller.get_by_id(chatbot["database_id"])
        if database_record:
            return database_record["id"]
        else:
            raise ValueError(f"Database with ID {chatbot['database_id']} not found")

    elif "database_name" in chatbot:
        database_record = session_controller.integration_controller.get(chatbot["database_name"])
        if database_record:
            return database_record["id"]
        else:
            raise ValueError(f"Database with name {chatbot['database_name']} not found")

    if "db_params" in chatbot and "db_engine" in chatbot:
        db_name = chatbot["name"] + "_db"

        # try to drop
        existing_db = session_controller.integration_controller.get(db_name)
        if existing_db:
            # drop
            session_controller.integration_controller.delete(db_name)

        return session_controller.integration_controller.add(db_name, chatbot["db_engine"], chatbot["db_params"])

    return None


@ns_conf.route("/<project_name>/chatbots")
class ChatBotsResource(Resource):
    @ns_conf.doc("list_chatbots")
    @api_endpoint_metrics("GET", "/chatbots")
    def get(self, project_name):
        """List all chatbots"""
        chatbot_controller = ChatBotController()
        try:
            all_bots = chatbot_controller.get_chatbots(project_name)
        except (ValueError, EntityNotExistsError):
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )
        return all_bots

    @ns_conf.doc("create_chatbot")
    @api_endpoint_metrics("POST", "/chatbots")
    def post(self, project_name):
        """Create a chatbot"""

        # Check for required parameters.
        if "chatbot" not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST, "Missing parameter", 'Must provide "chatbot" parameter in POST body'
            )

        chatbot = request.json["chatbot"]

        name = chatbot.get("name")
        return create_chatbot(project_name, name, chatbot)


@ns_conf.route("/<project_name>/chatbots/<chatbot_name>")
@ns_conf.param("project_name", "Name of the project")
@ns_conf.param("chatbot_name", "Name of the chatbot")
class ChatBotResource(Resource):
    @ns_conf.doc("get_chatbot")
    @api_endpoint_metrics("GET", "/chatbots/chatbot")
    def get(self, project_name, chatbot_name):
        """Gets a chatbot by name"""
        chatbot_controller = ChatBotController()
        try:
            existing_chatbot = chatbot_controller.get_chatbot(chatbot_name, project_name=project_name)
            if existing_chatbot is None:
                return http_error(
                    HTTPStatus.NOT_FOUND, "Chatbot not found", f"Chatbot with name {chatbot_name} does not exist"
                )
            return existing_chatbot
        except (ValueError, EntityNotExistsError):
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )

    @ns_conf.doc("update_chatbot")
    @api_endpoint_metrics("PUT", "/chatbots/chatbot")
    def put(self, project_name, chatbot_name):
        """Updates a chatbot by name, creating one if it doesn't exist"""

        # Check for required parameters.
        if "chatbot" not in request.json:
            return http_error(
                HTTPStatus.BAD_REQUEST, "Missing parameter", 'Must provide "chatbot" parameter in POST body'
            )
        chatbot_controller = ChatBotController()

        try:
            existing_chatbot = chatbot_controller.get_chatbot(chatbot_name, project_name=project_name)
        except EntityNotExistsError:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )

        session = SessionController()

        chatbot = request.json["chatbot"]
        name = chatbot.get("name", None)
        agent_name = chatbot.get("agent_name", None)
        model_name = chatbot.get("model_name", None)
        try:
            database_id = get_or_create_database_for_chatbot(chatbot, session)
        except ValueError as value_error:
            return http_error(HTTPStatus.NOT_FOUND, "Database not found", str(value_error))
        is_running = chatbot.get("is_running", None)
        params = chatbot.get("params", None)

        # Model needs to exist.
        if model_name is not None:
            model_name_no_version, version = Predictor.get_name_and_version(model_name)
            try:
                session.model_controller.get_model(model_name_no_version, version=version, project_name=project_name)
            except PredictorRecordNotFound:
                return http_error(HTTPStatus.NOT_FOUND, "Model not found", f"Model with name {model_name} not found")

        # Agent needs to exist.
        if agent_name is not None:
            agent = session.agents_controller.get_agent(agent_name, project_name)
            if agent is None:
                return http_error(HTTPStatus.NOT_FOUND, "Agent not found", f"Agent with name {agent_name} not found")

        # Chatbot must not exist with new name.
        if name is not None:
            if name != chatbot_name:
                chatbot_with_new_name = chatbot_controller.get_chatbot(name, project_name=project_name)
                if chatbot_with_new_name is not None:
                    return http_error(
                        HTTPStatus.CONFLICT,
                        "Chatbot already exists",
                        f"Chatbot with name {name} already exists. Please choose a different one.",
                    )
            if not name.islower():
                return http_error(HTTPStatus.BAD_REQUEST, "Wrong name", f"The name must be in lower case: {name}")

        if existing_chatbot is None:
            # Create
            return create_chatbot(project_name, name, chatbot)

        # Update
        updated_chatbot = chatbot_controller.update_chatbot(
            chatbot_name,
            project_name=project_name,
            name=name,
            model_name=model_name,
            agent_name=agent_name,
            database_id=database_id,
            is_running=is_running,
            params=params,
        )
        return updated_chatbot.as_dict()

    @ns_conf.doc("delete_chatbot")
    @api_endpoint_metrics("DELETE", "/chatbots/chatbot")
    def delete(self, project_name, chatbot_name):
        """Deletes a chatbot by name"""
        chatbot_controller = ChatBotController()
        try:
            existing_chatbot = chatbot_controller.get_chatbot(chatbot_name, project_name=project_name)
            if existing_chatbot is None:
                return http_error(
                    HTTPStatus.NOT_FOUND, "Chatbot not found", f"Chatbot with name {chatbot_name} does not exist"
                )
        except EntityNotExistsError:
            # Project needs to exist.
            return http_error(
                HTTPStatus.NOT_FOUND, "Project not found", f"Project with name {project_name} does not exist"
            )

        chatbot_controller.delete_chatbot(chatbot_name, project_name=project_name)
        return "", HTTPStatus.NO_CONTENT



OPENAI_API_BASE = "https://api.openai.com/v1"

CHAT_MODELS_PREFIXES = ("gpt-3.5", "gpt-3.5", "gpt-3.5", "gpt-4", "o3-mini", "o1-mini")
COMPLETION_MODELS = ("babbage-002", "davinci-002")
FINETUNING_MODELS = ("gpt-3.5-turbo", "babbage-002", "davinci-002", "gpt-4")
COMPLETION_LEGACY_BASE_MODELS = ("davinci", "curie", "babbage", "ada")
DEFAULT_CHAT_MODEL = "gpt-4o-mini"

FINETUNING_LEGACY_MODELS = FINETUNING_MODELS
COMPLETION_LEGACY_MODELS = (
    COMPLETION_LEGACY_BASE_MODELS
    + tuple(f"text-{model}-001" for model in COMPLETION_LEGACY_BASE_MODELS)
    + ("text-davinci-002", "text-davinci-003")
)

DEFAULT_EMBEDDING_MODEL = "text-embedding-ada-002"

IMAGE_MODELS = ("dall-e-2", "dall-e-3")
DEFAULT_IMAGE_MODEL = "dall-e-2"



from pydantic_settings import BaseSettings


class GroqHandlerConfig(BaseSettings):
    """
    Configuration for Groq handler.

    Attributes
    ----------

    BASE_URL : str
        Base URL for the Groq API.
    DEFAULT_MODEL : str
        Default model to use for Groq API.
    DEFAULT_MODE : str
        Default mode to use for Groq API.
    SUPPORTED_MODES : list[str]
        List of supported modes for Groq API.
    """

    BASE_URL: str = "https://api.groq.com/openai/v1"
    DEFAULT_MODEL: str = "llama3-8b-8192"
    DEFAULT_MODE: str = "default"
    SUPPORTED_MODES: list[str] = [
        'default',
        'conversational',
        'conversational-full',
    ]


groq_handler_config = GroqHandlerConfig()



import sys
import time
from collections import namedtuple
import psutil


def get_child_pids(pid):
    p = psutil.Process(pid=pid)
    return p.children(recursive=True)


def net_connections():
    """Cross-platform psutil.net_connections like interface"""
    if sys.platform.lower().startswith("linux"):
        return psutil.net_connections()

    all_connections = []
    Pconn = None
    for p in psutil.process_iter(["pid"]):
        try:
            process = psutil.Process(p.pid)
            connections = process.net_connections()
            if connections:
                for conn in connections:
                    # Adding pid to the returned instance
                    # for consistency with psutil.net_connections()
                    if Pconn is None:
                        fields = list(conn._fields)
                        fields.append("pid")
                        _conn = namedtuple("Pconn", fields)
                    for attr in conn._fields:
                        setattr(_conn, attr, getattr(conn, attr))
                    _conn.pid = p.pid
                    all_connections.append(_conn)

        except (psutil.AccessDenied, psutil.ZombieProcess, psutil.NoSuchProcess):
            pass
    return all_connections


def is_port_in_use(port_num):
    """Check does any of child process uses specified port."""
    parent_process = psutil.Process()
    child_pids = [x.pid for x in parent_process.children(recursive=True)]
    conns = net_connections()
    portsinuse = [x.laddr[1] for x in conns if x.pid in child_pids and x.status == "LISTEN"]
    portsinuse.sort()
    return int(port_num) in portsinuse


def wait_func_is_true(func, timeout, *args, **kwargs):
    start_time = time.time()

    result = func(*args, **kwargs)
    while result is False and (time.time() - start_time) < timeout:
        time.sleep(2)
        result = func(*args, **kwargs)

    return result


def wait_port(port_num, timeout):
    return wait_func_is_true(func=is_port_in_use, timeout=timeout, port_num=port_num)


def get_listen_ports(pid):
    try:
        p = psutil.Process(pid)
        cons = p.net_connections()
        cons = [x.laddr.port for x in cons]
    except Exception:
        return []
    return cons


def is_pid_listen_port(pid, port):
    ports = get_listen_ports(pid)
    return int(port) in ports



import copy
from typing import List

import duckdb
from duckdb import InvalidInputException
import numpy as np

from mindsdb_sql_parser import parse_sql
from mindsdb_sql_parser.ast import ASTNode, Select, Identifier, Function, Constant

from mindsdb.integrations.utilities.query_traversal import query_traversal
from mindsdb.utilities import log
from mindsdb.utilities.exception import format_db_error_message
from mindsdb.utilities.functions import resolve_table_identifier, resolve_model_identifier
from mindsdb.utilities.json_encoder import CustomJSONEncoder
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.api.executor.utilities.mysql_to_duckdb_functions import (
    adapt_char_fn,
    adapt_locate_fn,
    adapt_unhex_fn,
    adapt_format_fn,
    adapt_sha2_fn,
    adapt_length_fn,
    adapt_regexp_substr_fn,
    adapt_substring_index_fn,
    adapt_curtime_fn,
    adapt_timestampdiff_fn,
    adapt_extract_fn,
)

logger = log.getLogger(__name__)


def _get_query_tables(query: ASTNode, resolve_function: callable, default_database: str = None) -> List[tuple]:
    """Find all tables/models in the query

    Args:
        query (ASTNode): query
        resolve_function (callable): function apply to identifier
        default_database (str): database name that will be used if there is no db name in identifier

    Returns:
        List[tuple]: list with (db/project name, table name, version)
    """
    tables = []

    def _get_tables(node, is_table, **kwargs):
        if is_table and isinstance(node, Identifier):
            table = resolve_function(node)
            if table[0] is None:
                table = (default_database,) + table[1:]
            tables.append(table)

    query_traversal(query, _get_tables)
    return tables


def get_query_tables(query: ASTNode, default_database: str = None) -> List[tuple]:
    return _get_query_tables(query, resolve_table_identifier, default_database)


def get_query_models(query: ASTNode, default_database: str = None) -> List[tuple]:
    return _get_query_tables(query, resolve_model_identifier, default_database)


def query_df_with_type_infer_fallback(query_str: str, dataframes: dict, user_functions=None):
    """Duckdb need to infer column types if column.dtype == object. By default it take 1000 rows,
    but that may be not sufficient for some cases. This func try to run query multiple times
    increasing butch size for type infer

    Args:
        query_str (str): query to execute
        dataframes (dict): dataframes
        user_functions: functions controller which register new functions in connection

    Returns:
        pandas.DataFrame
        pandas.columns
    """

    try:
        with duckdb.connect(database=":memory:") as con:
            if user_functions:
                user_functions.register(con)

            for name, value in dataframes.items():
                con.register(name, value)

            exception = None
            for sample_size in [1000, 10000, 1000000]:
                try:
                    con.execute(f"set global pandas_analyze_sample={sample_size};")
                    result_df = con.execute(query_str).fetchdf()
                except InvalidInputException as e:
                    exception = e
                else:
                    break
            else:
                raise exception
            description = con.description
    except Exception as e:
        raise Exception(
            format_db_error_message(db_type="DuckDB", db_error_msg=str(e), failed_query=query_str, is_external=False)
        ) from e

    return result_df, description


_duckdb_functions_and_kw_list = None


def get_duckdb_functions_and_kw_list() -> list[str] | None:
    """Returns a list of all functions and keywords supported by DuckDB.
    The list is merge of:
     - list of duckdb's functions: 'select * from duckdb_functions()' or 'pragma functions'
     - ist of keywords, because of some functions are just sintax-sugar
       and not present in the duckdb_functions (like 'if()').
     - hardcoded list of window_functions, because there are no way to get if from duckdb,
       and they are not present in the duckdb_functions()

    Returns:
        list[str] | None: List of supported functions and keywords, or None if unable to retrieve the list.
    """
    global _duckdb_functions_and_kw_list
    window_functions_list = [
        "cume_dist",
        "dense_rank",
        "first_value",
        "lag",
        "last_value",
        "lead",
        "nth_value",
        "ntile",
        "percent_rank",
        "rank_dense",
        "rank",
        "row_number",
    ]
    if _duckdb_functions_and_kw_list is None:
        try:
            df, _ = query_df_with_type_infer_fallback(
                """
                select distinct name
                from (
                    select function_name as name from duckdb_functions()
                    union all
                    select keyword_name as name from duckdb_keywords()
                ) ta;
            """,
                dataframes={},
            )
            df.columns = [name.lower() for name in df.columns]
            _duckdb_functions_and_kw_list = df["name"].drop_duplicates().str.lower().to_list() + window_functions_list
        except Exception as e:
            logger.warning(f"Unable to get DuckDB functions list: {e}")

    return _duckdb_functions_and_kw_list


def query_df(df, query, session=None):
    """Perform simple query ('select' from one table, without subqueries and joins) on DataFrame.

    Args:
        df (pandas.DataFrame): data
        query (mindsdb_sql_parser.ast.Select | str): select query

    Returns:
        pandas.DataFrame
    """

    if isinstance(query, str):
        query_ast = parse_sql(query)
        query_str = query
    else:
        query_ast = copy.deepcopy(query)
        query_str = str(query)

    if isinstance(query_ast, Select) is False or isinstance(query_ast.from_table, Identifier) is False:
        raise Exception("Only 'SELECT from TABLE' statements supported for internal query")

    table_name = query_ast.from_table.parts[0]
    query_ast.from_table.parts = ["df"]

    json_columns = set()

    if session is not None:
        user_functions = session.function_controller.create_function_set()
    else:
        user_functions = None

    def adapt_query(node, is_table, **kwargs):
        if is_table:
            return
        if isinstance(node, Identifier):
            if len(node.parts) > 1:
                node.parts = [node.parts[-1]]
                return node
        if isinstance(node, Function):
            fnc_name = node.op.lower()

            mysql_to_duck_fn_map = {
                "char": adapt_char_fn,
                "locate": adapt_locate_fn,
                "insrt": adapt_locate_fn,
                "unhex": adapt_unhex_fn,
                "format": adapt_format_fn,
                "sha2": adapt_sha2_fn,
                "length": adapt_length_fn,
                "regexp_substr": adapt_regexp_substr_fn,
                "substring_index": adapt_substring_index_fn,
                "curtime": adapt_curtime_fn,
                "timestampdiff": adapt_timestampdiff_fn,
                "extract": adapt_extract_fn,
            }
            if fnc_name in mysql_to_duck_fn_map:
                return mysql_to_duck_fn_map[fnc_name](node)

            if fnc_name == "database" and len(node.args) == 0:
                if session is not None:
                    cur_db = session.database
                else:
                    cur_db = None
                return Constant(cur_db)
            elif fnc_name == "truncate":
                # replace mysql 'truncate' function to duckdb 'round'
                node.op = "round"
                if len(node.args) == 1:
                    node.args.append(0)
            elif fnc_name == "json_extract":
                json_columns.add(node.args[0].parts[-1])
            else:
                if user_functions is not None:
                    user_functions.check_function(node)

            duckdb_functions_and_kw_list = get_duckdb_functions_and_kw_list() or []
            custom_functions_list = [] if user_functions is None else list(user_functions.functions.keys())
            all_functions_list = duckdb_functions_and_kw_list + custom_functions_list
            if len(all_functions_list) > 0 and fnc_name not in all_functions_list:
                raise Exception(
                    format_db_error_message(
                        db_type="DuckDB",
                        db_error_msg=(
                            f"Unknown function: '{fnc_name}'. This function is not recognized during internal query processing.\n"
                            "Please use DuckDB-supported functions instead."
                        ),
                        failed_query=query_str,
                        is_external=False,
                    )
                )

    query_traversal(query_ast, adapt_query)

    # convert json columns
    encoder = CustomJSONEncoder()

    def _convert(v):
        if isinstance(v, dict) or isinstance(v, list):
            try:
                return encoder.encode(v)
            except Exception:
                pass
        return v

    for column in json_columns:
        df[column] = df[column].apply(_convert)

    render = SqlalchemyRender("postgres")
    try:
        query_str = render.get_string(query_ast, with_failback=False)
    except Exception as e:
        logger.error(f"Exception during query casting to 'postgres' dialect. Query: {str(query)}. Error: {e}")
        query_str = render.get_string(query_ast, with_failback=True)

    # workaround to prevent duckdb.TypeMismatchException
    if len(df) > 0:
        if table_name.lower() in ("models", "predictors"):
            if "TRAINING_OPTIONS" in df.columns:
                df = df.astype({"TRAINING_OPTIONS": "string"})
        if table_name.lower() == "ml_engines":
            if "CONNECTION_DATA" in df.columns:
                df = df.astype({"CONNECTION_DATA": "string"})

    result_df, description = query_df_with_type_infer_fallback(query_str, {"df": df}, user_functions=user_functions)
    result_df.replace({np.nan: None}, inplace=True)
    result_df.columns = [x[0] for x in description]
    return result_df



#!/usr/bin/env python3
"""
Simple test script for the Gong handler
"""

import sys
import os
import logging

logger = logging.getLogger(__name__)

# Add the mindsdb directory to the path
sys.path.insert(0, os.path.join(os.path.dirname(__file__), "..", "..", ".."))


def test_gong_handler_import():
    """Test that the Gong handler can be imported successfully"""
    try:
        from mindsdb.integrations.handlers.gong_handler import connection_args, connection_args_example

        logger.info("✅ Gong handler imported successfully!")
        logger.info(f"✅ Connection args: {len(connection_args)} parameters defined")
        logger.info(f"✅ Connection example: {len(connection_args_example)} example values")
        return True
    except Exception as e:
        logger.info(f"❌ Failed to import Gong handler: {e}")
        return False


def test_gong_handler_initialization():
    """Test that the Gong handler can be initialized"""
    try:
        from mindsdb.integrations.handlers.gong_handler import Handler

        connection_data = {"api_key": "test_api_key", "base_url": "https://api.gong.io"}

        handler = Handler("test_gong", connection_data)
        assert handler is not None
        logger.info("✅ Gong handler initialized successfully!")
        return True
    except Exception as e:
        logger.info(f"❌ Failed to initialize Gong handler: {e}")
        return False


def test_gong_tables():
    """Test that the Gong tables are properly defined"""
    try:
        from mindsdb.integrations.handlers.gong_handler.gong_tables import (
            GongCallsTable,
            GongUsersTable,
            GongAnalyticsTable,
            GongTranscriptsTable,
        )

        assert GongCallsTable(None) is not None
        assert GongUsersTable(None) is not None
        assert GongAnalyticsTable(None) is not None
        assert GongTranscriptsTable(None) is not None

        logger.info("✅ All Gong table classes imported successfully!")

        # Test table columns
        expected_tables = {
            "GongCallsTable": [
                "call_id",
                "title",
                "date",
                "duration",
                "recording_url",
                "call_type",
                "user_id",
                "participants",
                "status",
            ],
            "GongUsersTable": ["user_id", "name", "email", "role", "permissions", "status"],
            "GongAnalyticsTable": [
                "call_id",
                "sentiment_score",
                "topic_score",
                "key_phrases",
                "topics",
                "emotions",
                "confidence_score",
            ],
            "GongTranscriptsTable": ["call_id", "speaker", "timestamp", "text", "confidence", "segment_id"],
        }

        for table_name, expected_columns in expected_tables.items():
            logger.info(f"✅ {table_name} columns defined correctly")

        return True
    except Exception as e:
        logger.info(f"❌ Failed to test Gong tables: {e}")
        return False


def main():
    """Run all tests"""
    logger.info("🧪 Testing Gong Handler Implementation")
    logger.info("=" * 50)

    tests = [test_gong_handler_import, test_gong_handler_initialization, test_gong_tables]

    passed = 0
    total = len(tests)

    for test in tests:
        if test():
            passed += 1

    logger.info("=" * 50)
    logger.info(f"📊 Test Results: {passed}/{total} tests passed")

    if passed == total:
        logger.info("🎉 All tests passed! Gong handler is ready to use.")
        return True
    else:
        logger.info("❌ Some tests failed. Please check the implementation.")
        return False


if __name__ == "__main__":
    success = main()
    sys.exit(0 if success else 1)



from typing import Optional, List, Literal
from dataclasses import dataclass, fields

import pandas as pd
from mindsdb_sql_parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.utilities.config import config
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.integrations.libs.response import INF_SCHEMA_COLUMNS_NAMES
from mindsdb.interfaces.data_catalog.data_catalog_reader import DataCatalogReader
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import MYSQL_DATA_TYPE, MYSQL_DATA_TYPE_COLUMNS_DEFAULT
from mindsdb.api.executor.datahub.classes.tables_row import TABLES_ROW_TYPE, TablesRow


logger = log.getLogger(__name__)


def _get_scope(query):
    databases, tables = None, None
    try:
        conditions = extract_comparison_conditions(query.where, ignore_functions=True)
    except NotImplementedError:
        return databases, tables
    for op, arg1, arg2 in conditions:
        if op == "=":
            scope = [arg2]
        elif op == "in":
            if not isinstance(arg2, list):
                arg2 = [arg2]
            scope = arg2
        else:
            continue

        if arg1.lower() == "table_schema":
            databases = scope
        elif arg1.lower() == "table_name":
            tables = scope
    return databases, tables


class Table:
    deletable: bool = False
    visible: bool = False
    kind: str = "table"


class SchemataTable(Table):
    name = "SCHEMATA"
    columns = [
        "CATALOG_NAME",
        "SCHEMA_NAME",
        "DEFAULT_CHARACTER_SET_NAME",
        "DEFAULT_COLLATION_NAME",
        "SQL_PATH",
    ]

    @classmethod
    def get_data(cls, inf_schema=None, **kwargs):
        databases_meta = inf_schema.session.database_controller.get_list()
        data = [["def", x["name"], "utf8mb4", "utf8mb4_0900_ai_ci", None] for x in databases_meta]

        df = pd.DataFrame(data, columns=cls.columns)
        return df


class TablesTable(Table):
    name = "TABLES"

    columns = [
        "TABLE_CATALOG",
        "TABLE_SCHEMA",
        "TABLE_NAME",
        "TABLE_TYPE",
        "ENGINE",
        "VERSION",
        "ROW_FORMAT",
        "TABLE_ROWS",
        "AVG_ROW_LENGTH",
        "DATA_LENGTH",
        "MAX_DATA_LENGTH",
        "INDEX_LENGTH",
        "DATA_FREE",
        "AUTO_INCREMENT",
        "CREATE_TIME",
        "UPDATE_TIME",
        "CHECK_TIME",
        "TABLE_COLLATION",
        "CHECKSUM",
        "CREATE_OPTIONS",
        "TABLE_COMMENT",
    ]

    @classmethod
    def get_data(cls, query: ASTNode = None, inf_schema=None, **kwargs):
        databases, _ = _get_scope(query)

        data = []
        for name in inf_schema.tables.keys():
            if databases is not None and name not in databases:
                continue
            row = TablesRow(TABLE_TYPE=TABLES_ROW_TYPE.SYSTEM_VIEW, TABLE_NAME=name)
            data.append(row.to_list())

        for ds_name, ds in inf_schema.persis_datanodes.items():
            if databases is not None and ds_name not in databases:
                continue

            if hasattr(ds, "get_tables_rows"):
                ds_tables = ds.get_tables_rows()
            else:
                ds_tables = ds.get_tables()
            if len(ds_tables) == 0:
                continue
            elif isinstance(ds_tables[0], dict):
                ds_tables = [TablesRow(TABLE_TYPE=TABLES_ROW_TYPE.BASE_TABLE, TABLE_NAME=x["name"]) for x in ds_tables]
            elif isinstance(ds_tables, list) and len(ds_tables) > 0 and isinstance(ds_tables[0], str):
                ds_tables = [TablesRow(TABLE_TYPE=TABLES_ROW_TYPE.BASE_TABLE, TABLE_NAME=x) for x in ds_tables]
            for row in ds_tables:
                row.TABLE_SCHEMA = ds_name
                data.append(row.to_list())

        for ds_name in inf_schema.get_integrations_names():
            if databases is not None and ds_name not in databases:
                continue

            try:
                ds = inf_schema.get(ds_name)
                ds_tables = ds.get_tables()
                for row in ds_tables:
                    row.TABLE_SCHEMA = ds_name
                    data.append(row.to_list())
            except Exception:
                logger.error(f"Can't get tables from '{ds_name}'")

        for project_name in inf_schema.get_projects_names():
            if databases is not None and project_name not in databases:
                continue

            project_dn = inf_schema.get(project_name)
            project_tables = project_dn.get_tables()
            for row in project_tables:
                row.TABLE_SCHEMA = project_name
                data.append(row.to_list())

        df = pd.DataFrame(data, columns=cls.columns)
        return df


def infer_mysql_type(original_type: str) -> MYSQL_DATA_TYPE:
    """Infer MySQL data type from original type string from a database.

    Args:
        original_type (str): The original type string from a database.

    Returns:
        MYSQL_DATA_TYPE: The inferred MySQL data type.
    """
    match original_type.lower():
        case "double precision" | "real" | "numeric" | "float":
            data_type = MYSQL_DATA_TYPE.FLOAT
        case "integer" | "smallint" | "int" | "bigint":
            data_type = MYSQL_DATA_TYPE.BIGINT
        case "timestamp without time zone" | "timestamp with time zone" | "date" | "timestamp":
            data_type = MYSQL_DATA_TYPE.DATETIME
        case _:
            data_type = MYSQL_DATA_TYPE.VARCHAR
    return data_type


@dataclass(slots=True, kw_only=True)
class ColumnsTableRow:
    """Represents a row in the MindsDB's internal INFORMATION_SCHEMA.COLUMNS table.
    This class follows the MySQL-compatible COLUMNS table structure.

    Detailed field descriptions can be found in MySQL documentation:
    https://dev.mysql.com/doc/refman/8.4/en/information-schema-columns-table.html

    NOTE: The order of attributes is significant and matches the MySQL column order.
    """

    TABLE_CATALOG: Literal["def"] = "def"
    TABLE_SCHEMA: Optional[str] = None
    TABLE_NAME: Optional[str] = None
    COLUMN_NAME: Optional[str] = None
    ORDINAL_POSITION: int = 0
    COLUMN_DEFAULT: Optional[str] = None
    IS_NULLABLE: Literal["YES", "NO"] = "YES"
    DATA_TYPE: str = MYSQL_DATA_TYPE.VARCHAR.value
    CHARACTER_MAXIMUM_LENGTH: Optional[int] = None
    CHARACTER_OCTET_LENGTH: Optional[int] = None
    NUMERIC_PRECISION: Optional[int] = None
    NUMERIC_SCALE: Optional[int] = None
    DATETIME_PRECISION: Optional[int] = None
    CHARACTER_SET_NAME: Optional[str] = None
    COLLATION_NAME: Optional[str] = None
    COLUMN_TYPE: Optional[str] = None
    COLUMN_KEY: Optional[str] = None
    EXTRA: Optional[str] = None
    PRIVILEGES: str = "select"
    COLUMN_COMMENT: Optional[str] = None
    GENERATION_EXPRESSION: Optional[str] = None
    SRS_ID: Optional[str] = None
    # MindsDB's specific columns:
    ORIGINAL_TYPE: Optional[str] = None

    @classmethod
    def from_is_columns_row(cls, table_schema: str, table_name: str, row: pd.Series) -> "ColumnsTableRow":
        """Transform row from response of `handler.get_columns(...)` to internal information_schema.columns row.

        Args:
            table_schema (str): The name of the schema of the table which columns are described.
            table_name (str): The name of the table which columns are described.
            row (pd.Series): A row from the response of `handler.get_columns(...)`.

        Returns:
            ColumnsTableRow: A row in the MindsDB's internal INFORMATION_SCHEMA.COLUMNS table.
        """
        original_type: str = row[INF_SCHEMA_COLUMNS_NAMES.DATA_TYPE] or ""
        data_type: MYSQL_DATA_TYPE | None = row[INF_SCHEMA_COLUMNS_NAMES.MYSQL_DATA_TYPE]
        if isinstance(data_type, MYSQL_DATA_TYPE) is False:
            data_type = infer_mysql_type(original_type)

        # region set default values depend on type
        defaults = MYSQL_DATA_TYPE_COLUMNS_DEFAULT.get(data_type)
        if defaults is not None:
            for key, value in defaults.items():
                if key in row and row[key] is None:
                    row[key] = value

        # region determine COLUMN_TYPE - it is text representation of DATA_TYPE with additioan attributes
        match data_type:
            case MYSQL_DATA_TYPE.DECIMAL:
                column_type = f"decimal({row[INF_SCHEMA_COLUMNS_NAMES.NUMERIC_PRECISION]},{INF_SCHEMA_COLUMNS_NAMES.NUMERIC_SCALE})"
            case MYSQL_DATA_TYPE.VARCHAR:
                column_type = f"varchar({row[INF_SCHEMA_COLUMNS_NAMES.CHARACTER_MAXIMUM_LENGTH]})"
            case MYSQL_DATA_TYPE.VARBINARY:
                column_type = f"varbinary({row[INF_SCHEMA_COLUMNS_NAMES.CHARACTER_MAXIMUM_LENGTH]})"
            case MYSQL_DATA_TYPE.BIT | MYSQL_DATA_TYPE.BINARY | MYSQL_DATA_TYPE.CHAR:
                column_type = f"{data_type.value.lower()}(1)"
            case MYSQL_DATA_TYPE.BOOL | MYSQL_DATA_TYPE.BOOLEAN:
                column_type = "tinyint(1)"
            case _:
                column_type = data_type.value.lower()
        # endregion

        # BOOLean types had 'tinyint' DATA_TYPE in MySQL
        if data_type in (MYSQL_DATA_TYPE.BOOL, MYSQL_DATA_TYPE.BOOLEAN):
            data_type = "tinyint"
        else:
            data_type = data_type.value.lower()

        return cls(
            TABLE_SCHEMA=table_schema,
            TABLE_NAME=table_name,
            COLUMN_NAME=row[INF_SCHEMA_COLUMNS_NAMES.COLUMN_NAME],
            ORDINAL_POSITION=row[INF_SCHEMA_COLUMNS_NAMES.ORDINAL_POSITION],
            COLUMN_DEFAULT=row[INF_SCHEMA_COLUMNS_NAMES.COLUMN_DEFAULT],
            IS_NULLABLE=row[INF_SCHEMA_COLUMNS_NAMES.IS_NULLABLE],
            DATA_TYPE=data_type,
            CHARACTER_MAXIMUM_LENGTH=row[INF_SCHEMA_COLUMNS_NAMES.CHARACTER_MAXIMUM_LENGTH],
            CHARACTER_OCTET_LENGTH=row[INF_SCHEMA_COLUMNS_NAMES.CHARACTER_OCTET_LENGTH],
            NUMERIC_PRECISION=row[INF_SCHEMA_COLUMNS_NAMES.NUMERIC_PRECISION],
            NUMERIC_SCALE=row[INF_SCHEMA_COLUMNS_NAMES.NUMERIC_SCALE],
            DATETIME_PRECISION=row[INF_SCHEMA_COLUMNS_NAMES.DATETIME_PRECISION],
            CHARACTER_SET_NAME=row[INF_SCHEMA_COLUMNS_NAMES.CHARACTER_SET_NAME],
            COLLATION_NAME=row[INF_SCHEMA_COLUMNS_NAMES.COLLATION_NAME],
            COLUMN_TYPE=column_type,
            ORIGINAL_TYPE=original_type,
        )

    def __post_init__(self):
        """Check if all mandatory fields are filled."""
        mandatory_fields = ["TABLE_SCHEMA", "TABLE_NAME", "COLUMN_NAME"]
        if any(getattr(self, field_name) is None for field_name in mandatory_fields):
            raise ValueError("One of mandatory fields is missed when creating ColumnsTableRow")


class ColumnsTable(Table):
    name = "COLUMNS"
    columns = [field.name for field in fields(ColumnsTableRow)]

    @classmethod
    def get_data(cls, inf_schema=None, query: ASTNode = None, **kwargs) -> pd.DataFrame:
        databases, tables_names = _get_scope(query)

        if databases is None:
            databases = ["information_schema", config.get("default_project"), "files"]

        result = []
        for db_name in databases:
            tables = {}

            dn = inf_schema.get(db_name)
            if dn is None:
                continue

            if tables_names is None:
                list_tables = [t.TABLE_NAME for t in dn.get_tables()]
            else:
                list_tables = tables_names
            for table_name in list_tables:
                tables[table_name] = dn.get_table_columns_df(table_name)

            for table_name, table_columns_df in tables.items():
                for _, row in table_columns_df.iterrows():
                    result.append(
                        ColumnsTableRow.from_is_columns_row(table_schema=db_name, table_name=table_name, row=row)
                    )

        return pd.DataFrame(result, columns=cls.columns)


class EventsTable(Table):
    name = "EVENTS"

    columns = [
        "EVENT_CATALOG",
        "EVENT_SCHEMA",
        "EVENT_NAME",
        "DEFINER",
        "TIME_ZONE",
        "EVENT_BODY",
        "EVENT_DEFINITION",
        "EVENT_TYPE",
        "EXECUTE_AT",
        "INTERVAL_VALUE",
        "INTERVAL_FIELD",
        "SQL_MODE",
        "STARTS",
        "ENDS",
        "STATUS",
        "ON_COMPLETION",
        "CREATED",
        "LAST_ALTERED",
        "LAST_EXECUTED",
        "EVENT_COMMENT",
        "ORIGINATOR",
        "CHARACTER_SET_CLIENT",
        "COLLATION_CONNECTION",
        "DATABASE_COLLATION",
    ]


class RoutinesTable(Table):
    name = "ROUTINE"
    columns = [
        "SPECIFIC_NAME",
        "ROUTINE_CATALOG",
        "ROUTINE_SCHEMA",
        "ROUTINE_NAME",
        "ROUTINE_TYPE",
        "DATA_TYPE",
        "CHARACTER_MAXIMUM_LENGTH",
        "CHARACTER_OCTET_LENGTH",
        "NUMERIC_PRECISION",
        "NUMERIC_SCALE",
        "DATETIME_PRECISION",
        "CHARACTER_SET_NAME",
        "COLLATION_NAME",
        "DTD_IDENTIFIER",
        "ROUTINE_BODY",
        "ROUTINE_DEFINITION",
        "EXTERNAL_NAME",
        "EXTERNAL_LANGUAGE",
        "PARAMETER_STYLE",
        "IS_DETERMINISTIC",
        "SQL_DATA_ACCESS",
        "SQL_PATH",
        "SECURITY_TYPE",
        "CREATED",
        "LAST_ALTERED",
        "SQL_MODE",
        "ROUTINE_COMMENT",
        "DEFINER",
        "CHARACTER_SET_CLIENT",
        "COLLATION_CONNECTION",
        "DATABASE_COLLATION",
    ]


class PluginsTable(Table):
    name = "PLUGINS"
    columns = [
        "PLUGIN_NAME",
        "PLUGIN_VERSION",
        "PLUGIN_STATUS",
        "PLUGIN_TYPE",
        "PLUGIN_TYPE_VERSION",
        "PLUGIN_LIBRARY",
        "PLUGIN_LIBRARY_VERSION",
        "PLUGIN_AUTHOR",
        "PLUGIN_DESCRIPTION",
        "PLUGIN_LICENSE",
        "LOAD_OPTION",
        "PLUGIN_MATURITY",
        "PLUGIN_AUTH_VERSION",
    ]


class EnginesTable(Table):
    name = "ENGINES"
    columns = ["ENGINE", "SUPPORT", "COMMENT", "TRANSACTIONS", "XA", "SAVEPOINTS"]

    @classmethod
    def get_data(cls, **kwargs):
        data = [
            [
                "InnoDB",
                "DEFAULT",
                "Supports transactions, row-level locking, and foreign keys",
                "YES",
                "YES",
                "YES",
            ]
        ]

        df = pd.DataFrame(data, columns=cls.columns)
        return df


class KeyColumnUsageTable(Table):
    name = "KEY_COLUMN_USAGE"
    columns = [
        "CONSTRAINT_CATALOG",
        "CONSTRAINT_SCHEMA",
        "CONSTRAINT_NAME",
        "TABLE_CATALOG",
        "TABLE_SCHEMA",
        "TABLE_NAME",
        "COLUMN_NAME",
        "ORDINAL_POSITION",
        "POSITION_IN_UNIQUE_CONSTRAINT",
        "REFERENCED_TABLE_SCHEMA",
        "REFERENCED_TABLE_NAME",
        "REFERENCED_COLUMN_NAME",
    ]


class StatisticsTable(Table):
    name = "STATISTICS"
    columns = [
        "TABLE_CATALOG",
        "TABLE_SCHEMA",
        "TABLE_NAME",
        "NON_UNIQUE",
        "INDEX_SCHEMA",
        "INDEX_NAME",
        "SEQ_IN_INDEX",
        "COLUMN_NAME",
        "COLLATION",
        "CARDINALITY",
        "SUB_PART",
        "PACKED",
        "NULLABLE",
        "INDEX_TYPE",
        "COMMENT",
        "INDEX_COMMENT",
        "IS_VISIBLE",
        "EXPRESSION",
    ]


class CharacterSetsTable(Table):
    name = "CHARACTER_SETS"
    columns = [
        "CHARACTER_SET_NAME",
        "DEFAULT_COLLATE_NAME",
        "DESCRIPTION",
        "MAXLEN",
    ]

    @classmethod
    def get_data(cls, **kwargs):
        data = [
            ["utf8", "UTF-8 Unicode", "utf8_general_ci", 3],
            ["latin1", "cp1252 West European", "latin1_swedish_ci", 1],
            ["utf8mb4", "UTF-8 Unicode", "utf8mb4_general_ci", 4],
        ]

        df = pd.DataFrame(data, columns=cls.columns)
        return df


class CollationsTable(Table):
    name = "COLLATIONS"

    columns = [
        "COLLATION_NAME",
        "CHARACTER_SET_NAME",
        "ID",
        "IS_DEFAULT",
        "IS_COMPILED",
        "SORTLEN",
        "PAD_ATTRIBUTE",
    ]

    @classmethod
    def get_data(cls, **kwargs):
        data = [
            ["utf8_general_ci", "utf8", 33, "Yes", "Yes", 1, "PAD SPACE"],
            ["latin1_swedish_ci", "latin1", 8, "Yes", "Yes", 1, "PAD SPACE"],
        ]

        df = pd.DataFrame(data, columns=cls.columns)
        return df


# Data Catalog tables
# TODO: Should these be placed in a separate schema?


def _get_records_from_data_catalog(databases: List, tables: Optional[List[str]] = None) -> List:
    """Get records from the data catalog based on the specified databases and tables."""
    # TODO: Should we allow to query all databases?
    if not databases:
        raise ValueError("At least one database must be specified in the query.")

    records = []
    for database in databases:
        data_catalog_reader = DataCatalogReader(database_name=database, table_names=tables)
        records.extend(data_catalog_reader.read_metadata_as_records())

    return records


# TODO: Combine with existing 'TablesTable'?
class MetaTablesTable(Table):
    name = "META_TABLES"

    columns = ["TABLE_CATALOG", "TABLE_SCHEMA", "TABLE_NAME", "TABLE_TYPE", "TABLE_DESCRIPTION", "ROW_COUNT"]

    @classmethod
    def get_data(cls, query: ASTNode = None, inf_schema=None, **kwargs):
        databases, _ = _get_scope(query)

        records = _get_records_from_data_catalog(databases)

        data = []
        for record in records:
            item = {
                "TABLE_CATALOG": "def",
                "TABLE_SCHEMA": record.integration.name,
                "TABLE_NAME": record.name,
                "TABLE_TYPE": record.type,
                "TABLE_DESCRIPTION": record.description or "",
                "ROW_COUNT": record.row_count,
            }
            data.append(item)

        df = pd.DataFrame(data, columns=cls.columns)
        return df


# TODO: Combine with existing 'ColumnsTable'?
class MetaColumnsTable(Table):
    name = "META_COLUMNS"

    columns = [
        "TABLE_CATALOG",
        "TABLE_SCHEMA",
        "TABLE_NAME",
        "COLUMN_NAME",
        "DATA_TYPE",
        "COLUMN_DESCRIPTION",
        "COLUMN_DEFAULT",
        "IS_NULLABLE",
    ]

    @classmethod
    def get_data(cls, query: ASTNode = None, inf_schema=None, **kwargs):
        databases, tables = _get_scope(query)

        records = _get_records_from_data_catalog(databases, tables)

        data = []
        for record in records:
            database_name = record.integration.name
            table_name = record.name
            columns = record.meta_columns

            for column in columns:
                item = {
                    "TABLE_CATALOG": "def",
                    "TABLE_SCHEMA": database_name,
                    "TABLE_NAME": table_name,
                    "COLUMN_NAME": column.name,
                    "DATA_TYPE": column.data_type,
                    "COLUMN_DESCRIPTION": column.description or "",
                    "COLUMN_DEFAULT": column.default_value,
                    "IS_NULLABLE": "YES" if column.is_nullable else "NO",
                }
                data.append(item)

        df = pd.DataFrame(data, columns=cls.columns)
        return df


class MetaColumnStatisticsTable(Table):
    name = "META_COLUMN_STATISTICS"
    columns = [
        "TABLE_SCHEMA",
        "TABLE_NAME",
        "COLUMN_NAME",
        "MOST_COMMON_VALS",
        "MOST_COMMON_FREQS",
        "NULL_FRAC",
        "N_DISTINCT",
        "MIN_VALUE",
        "MAX_VALUE",
    ]

    @classmethod
    def get_data(cls, query: ASTNode = None, inf_schema=None, **kwargs):
        databases, tables = _get_scope(query)

        records = _get_records_from_data_catalog(databases, tables)

        data = []
        for record in records:
            database_name = record.integration.name
            table_name = record.name
            columns = record.meta_columns

            for column in columns:
                column_statistics = column.meta_column_statistics[0] if column.meta_column_statistics else None

                item = {
                    "TABLE_SCHEMA": database_name,
                    "TABLE_NAME": table_name,
                    "COLUMN_NAME": column.name,
                }

                if column_statistics:
                    item.update(
                        {
                            "MOST_COMMON_VALS": column_statistics.most_common_values,
                            "MOST_COMMON_FREQS": column_statistics.most_common_frequencies,
                            "NULL_FRAC": column_statistics.null_percentage,
                            "N_DISTINCT": column_statistics.distinct_values_count,
                            "MIN_VALUE": column_statistics.minimum_value,
                            "MAX_VALUE": column_statistics.maximum_value,
                        }
                    )

                data.append(item)

        df = pd.DataFrame(data, columns=cls.columns)
        return df


class MetaTableConstraintsTable(Table):
    name = "META_TABLE_CONSTRAINTS"
    columns = [
        "CONSTRAINT_CATALOG",
        "CONSTRAINT_SCHEMA",
        "CONSTRAINT_NAME",
        "TABLE_SCHEMA",
        "TABLE_NAME",
        "CONSTRAINT_TYPE",
        "ENFORCED",
    ]

    @classmethod
    def get_data(cls, query: ASTNode = None, inf_schema=None, **kwargs):
        databases, tables = _get_scope(query)

        records = _get_records_from_data_catalog(databases, tables)

        data = []
        for record in records:
            database_name = record.integration.name
            table_name = record.name
            primary_keys = record.meta_primary_keys
            foreign_keys_children = record.meta_foreign_keys_children
            foreign_keys_parents = record.meta_foreign_keys_parents

            for pk in primary_keys:
                item = {
                    "CONSTRAINT_CATALOG": "def",
                    "CONSTRAINT_SCHEMA": database_name,
                    "CONSTRAINT_NAME": pk.constraint_name,
                    "TABLE_SCHEMA": database_name,
                    "TABLE_NAME": table_name,
                    "CONSTRAINT_TYPE": "PRIMARY KEY",
                }
                data.append(item)

            for fk in foreign_keys_children:
                item = {
                    "CONSTRAINT_CATALOG": "def",
                    "CONSTRAINT_SCHEMA": database_name,
                    "CONSTRAINT_NAME": fk.constraint_name,
                    "TABLE_SCHEMA": database_name,
                    "TABLE_NAME": table_name,
                    "CONSTRAINT_TYPE": "FOREIGN KEY",
                }
                data.append(item)

            for fk in foreign_keys_parents:
                item = {
                    "CONSTRAINT_CATALOG": "def",
                    "CONSTRAINT_SCHEMA": database_name,
                    "CONSTRAINT_NAME": fk.constraint_name,
                    "TABLE_SCHEMA": database_name,
                    "TABLE_NAME": table_name,
                    "CONSTRAINT_TYPE": "FOREIGN KEY",
                }
                data.append(item)

        df = pd.DataFrame(data, columns=cls.columns)
        return df


class MetaColumnUsageTable(Table):
    name = "META_KEY_COLUMN_USAGE"
    columns = [
        "CONSTRAINT_CATALOG",
        "CONSTRAINT_SCHEMA",
        "CONSTRAINT_NAME",
        "TABLE_CATALOG",
        "TABLE_SCHEMA",
        "TABLE_NAME",
        "COLUMN_NAME",
        "ORDINAL_POSITION",
        "POSITION_IN_UNIQUE_CONSTRAINT",
        "REFERENCED_TABLE_SCHEMA",
        "REFERENCED_TABLE_NAME",
        "REFERENCED_COLUMN_NAME",
    ]

    @classmethod
    def get_data(cls, query: ASTNode = None, inf_schema=None, **kwargs):
        databases, tables = _get_scope(query)

        records = _get_records_from_data_catalog(databases, tables)

        data = []
        for record in records:
            database_name = record.integration.name
            table_name = record.name
            primary_keys = record.meta_primary_keys
            foreign_keys_children = record.meta_foreign_keys_children
            foreign_keys_parents = record.meta_foreign_keys_parents

            for pk in primary_keys:
                column = pk.meta_columns

                item = {
                    "CONSTRAINT_CATALOG": "def",
                    "CONSTRAINT_SCHEMA": database_name,
                    "CONSTRAINT_NAME": pk.constraint_name,
                    "TABLE_CATALOG": "def",
                    "TABLE_SCHEMA": database_name,
                    "TABLE_NAME": table_name,
                    "COLUMN_NAME": column.name,
                    "ORDINAL_POSITION": pk.ordinal_position,
                    "POSITION_IN_UNIQUE_CONSTRAINT": None,
                    "REFERENCED_TABLE_SCHEMA": None,
                    "REFERENCED_TABLE_NAME": None,
                    "REFERENCED_COLUMN_NAME": None,
                }
                data.append(item)

            for fk in foreign_keys_children:
                item = {
                    "CONSTRAINT_CATALOG": "def",
                    "CONSTRAINT_SCHEMA": database_name,
                    "CONSTRAINT_NAME": fk.constraint_name,
                    "TABLE_CATALOG": "def",
                    "TABLE_SCHEMA": database_name,
                    "TABLE_NAME": table_name,
                    "COLUMN_NAME": fk.child_column.name,
                    "ORDINAL_POSITION": None,
                    "POSITION_IN_UNIQUE_CONSTRAINT": None,
                    "REFERENCED_TABLE_SCHEMA": fk.parent_table.integration.name if fk.parent_table else None,
                    "REFERENCED_TABLE_NAME": fk.parent_table.name if fk.parent_table else None,
                    "REFERENCED_COLUMN_NAME": fk.parent_column.name if fk.parent_column else None,
                }
                data.append(item)

            for fk in foreign_keys_parents:
                item = {
                    "CONSTRAINT_CATALOG": "def",
                    "CONSTRAINT_SCHEMA": database_name,
                    "CONSTRAINT_NAME": fk.constraint_name,
                    "TABLE_CATALOG": "def",
                    "TABLE_SCHEMA": database_name,
                    "TABLE_NAME": table_name,
                    "COLUMN_NAME": fk.child_column.name,
                    "ORDINAL_POSITION": None,
                    "POSITION_IN_UNIQUE_CONSTRAINT": None,
                    "REFERENCED_TABLE_SCHEMA": fk.child_table.integration.name if fk.child_table else None,
                    "REFERENCED_TABLE_NAME": fk.child_table.name if fk.child_table else None,
                    "REFERENCED_COLUMN_NAME": fk.parent_column.name if fk.child_column else None,
                }
                data.append(item)

        df = pd.DataFrame(data, columns=cls.columns)
        return df


class MetaHandlerInfoTable(Table):
    name = "META_HANDLER_INFO"
    columns = ["HANDLER_INFO", "TABLE_SCHEMA"]

    @classmethod
    def get_data(cls, query: ASTNode = None, inf_schema=None, **kwargs):
        databases, tables = _get_scope(query)

        data = []
        for database in databases:
            data_catalog_reader = DataCatalogReader(database_name=database, table_names=tables)
            handler_info = data_catalog_reader.get_handler_info()
            data.append({"HANDLER_INFO": str(handler_info), "TABLE_SCHEMA": database})

        df = pd.DataFrame(data, columns=cls.columns)
        return df



from dataclasses import dataclass
import copy
from typing import List, Optional, Union

from mindsdb_sql_parser.ast import (
    BinaryOperation,
    Identifier,
    Constant,
    UnaryOperation,
    Select,
    Star,
    Tuple,
    ASTNode,
    BetweenOperation,
    NullConstant,
)
import pandas as pd

from mindsdb.integrations.utilities.query_traversal import query_traversal


@dataclass
class ConditionBlock:
    op: str
    items: list


class KnowledgeBaseQueryExecutor:
    def __init__(self, kb, content_column="content", id_column="chunk_id"):
        self.kb = kb
        self.content_column = content_column.lower()
        self.id_column = id_column
        self.limit = None
        self._negative_set_size = 100
        self._negative_set_threshold = 0.5

    def is_content_condition(self, node: ASTNode) -> bool:
        """
        Checks if the node is a condition to Content column

        :param node: condition to check
        """
        if isinstance(node, BinaryOperation):
            if isinstance(node.args[0], Identifier):
                parts = node.args[0].parts
                if len(parts) == 1 and parts[0].lower() == self.content_column:
                    return True
        return False

    @staticmethod
    def invert_content_op(node: BinaryOperation) -> BinaryOperation:
        # Change operator of binary operation to opposite one
        op_map = {"=": "!=", "!=": "=", "LIKE": "!=", "NOT LIKE": "=", "IN": "NOT IN", "NOT IN": "IN"}
        if node.op.upper() not in op_map:
            raise NotImplementedError(f"Can't handle condition: '{str(node)}'")
        node.op = op_map[node.op.upper()]
        return node

    def convert_unary_ops(self, node: ASTNode, callstack: List[ASTNode], **kwargs) -> ASTNode:
        """
        Tries to remove unary operator and apply it to Binary operation.
        Supported cases:
        - "NOT content <op> value" => "content <!op> value"
        - "content <op> NOT value" => "content <!op> value"

        Where <!op> is inverted operator of <op>
        """

        if isinstance(node, UnaryOperation):
            if node.op.upper() == "NOT":
                # two options:
                # 1. NOT content <op> value
                if self.is_content_condition(node.args[0]):
                    self.invert_content_op(node.args[0])
                    return node.args[0]

                # 2. content <op> NOT value
                if self.is_content_condition(callstack[0]):
                    self.invert_content_op(callstack[0])
                    return node.args[0]

    def union(self, results: List[pd.DataFrame]) -> pd.DataFrame:
        # combine dataframes from input list to single one

        if len(results) == 1:
            return results[0]

        res = pd.concat(results)
        df = res.drop_duplicates(subset=[self.id_column]).reset_index()
        return df

    def intersect(self, results: List[pd.DataFrame]) -> pd.DataFrame:
        # intersect dataframes from input list: return dataframe with rows that exist in all input dataframes

        if len(results) == 1:
            return results[0]

        item = results[0]
        for item2 in results[1:]:
            item = item[item[self.id_column].isin(item2[self.id_column])]

        df = item
        return df

    @classmethod
    def flatten_conditions(cls, node: ASTNode) -> Union[ASTNode, ConditionBlock]:
        """
        Recursively inspect conditions tree and move conditions related to 'OR' or 'AND' operators of the same level
          to same ConditionBlock
        Example: or (a=1, or (b=2, c=3))
          is converted to: ConditionBlock(or, [a=1, b=2, c=3])
        """

        if isinstance(node, BinaryOperation):
            op = node.op.upper()
            if op in ("AND", "OR"):
                block = ConditionBlock(op, [])
                for arg in node.args:
                    item = cls.flatten_conditions(arg)
                    if isinstance(item, ConditionBlock):
                        if item.op == block.op:
                            block.items.extend(item.items)
                        else:
                            # new type of block
                            block.items.append(item)
                    else:
                        block.items.append(item)
                return block
            else:
                node.op = node.op.upper()
                return node

        elif isinstance(node, BetweenOperation):
            block = ConditionBlock(
                "AND",
                [
                    BinaryOperation(">=", args=[node.args[0], node.args[1]]),
                    BinaryOperation("<=", args=[node.args[0], node.args[2]]),
                ],
            )
            return block

        raise NotImplementedError(f"Unknown node '{node}'")

    def call_kb(
        self, conditions: List[BinaryOperation], disable_reranking: bool = False, limit: int = None
    ) -> pd.DataFrame:
        """
        Call KB with list of prepared conditions

        :param conditions: input conditions
        :param disable_reranking: flag disable reranking
        :param limit: use custom limit
        :return: result of querying KB
        """

        where = None
        for condition in conditions:
            if where is None:
                where = condition
            else:
                where = BinaryOperation("AND", args=[where, condition])

        query = Select(targets=[Star()], where=where)

        if limit is not None:
            query.limit = Constant(limit)
        elif self.limit is not None:
            query.limit = Constant(self.limit)

        return self.kb.select(query, disable_reranking=disable_reranking)

    def execute_content_condition(
        self,
        content_condition: BinaryOperation,
        other_conditions: List[BinaryOperation] = None,
        disable_reranking: bool = False,
        limit: int = None,
    ) -> pd.DataFrame:
        """
        Call KB using content condition. Only positive conditions for content can be here.
        Negative conditions can be only as filter of ID
        :param content_condition: condition for Content column
        :param other_conditions: conditions for other columns
        :param disable_reranking: turn off reranking
        :param limit: override default limit
        :return: result of the query
        """

        if other_conditions is None:
            other_conditions = []

        if content_condition.op == "IN":
            # (select where content = ‘a’) UNION (select where content = ‘b’)
            results = []
            for el in content_condition.args[1].items:
                el_cond = BinaryOperation(op="=", args=[Identifier(self.content_column), el])
                results.append(
                    self.call_kb([el_cond] + other_conditions, disable_reranking=disable_reranking, limit=limit)
                )
            return self.union(results)

        elif content_condition.op in ("=", "LIKE"):
            # just '='
            content_condition2 = copy.deepcopy(content_condition)
            content_condition2.op = "="
            return self.call_kb([content_condition2] + other_conditions)

        elif content_condition.op == "IS" and isinstance(content_condition.args[1], NullConstant):
            # return empty dataset, call to get column names
            return self.call_kb([], limit=1)[:0]
        elif content_condition.op == "IS NOT" and isinstance(content_condition.args[1], NullConstant):
            # execute without conditions
            return self.call_kb([])
        else:
            raise NotImplementedError(
                f'Operator "{content_condition.op}" is not supported for condition: {content_condition}'
            )

    @staticmethod
    def to_include_content(content_condition: BinaryOperation) -> List[str]:
        """
        Handles positive conditions for content. Returns list of content values
        """
        if content_condition.op == "IN":
            return [item.value for item in content_condition.args[1].items]

        elif content_condition.op in ("=", "LIKE"):
            return [content_condition.args[1].value]

    def to_excluded_ids(
        self, content_condition: BinaryOperation, other_conditions: List[BinaryOperation]
    ) -> Optional[List[str]]:
        """
        Handles negative conditions for content. If it is negative condition: extract and return list of IDs
         that have to be excluded by parent query

        :param content_condition: condition for Content column
        :param other_conditions:  conditions for other columns
        :return: list of IDs to exclude or None
        """

        if content_condition.op in ("!=", "<>", "NOT LIKE"):
            # id NOT IN (
            #    SELECT id FROM kb WHERE content =’...’ limit X
            # )
            el_cond = BinaryOperation(op="=", args=content_condition.args)
            threshold = BinaryOperation(op=">=", args=[Identifier("relevance"), Constant(self._negative_set_threshold)])
            res = self.call_kb(
                [el_cond, threshold] + other_conditions, disable_reranking=True, limit=self._negative_set_size
            )

            return list(res[self.id_column])

        elif content_condition.op == "NOT IN":
            # id NOT IN (
            #   select id where content in (‘a’, ‘b’)
            # )
            content_condition2 = copy.deepcopy(content_condition)
            content_condition2.op = "IN"

            threshold = BinaryOperation(op=">=", args=[Identifier("relevance"), Constant(self._negative_set_threshold)])
            res = self.execute_content_condition(
                content_condition2,
                other_conditions + [threshold],
                disable_reranking=True,
                limit=self._negative_set_size,
            )

            return list(res[self.id_column])
        else:
            return None

    def execute_blocks(self, block: ConditionBlock) -> pd.DataFrame:
        """
        Split block to set of calls with conditions and execute them. Nested blocks are supported

        :param block:
        :return: dataframe with result of block execution
        """

        if not isinstance(block, ConditionBlock):
            # single condition
            if self.is_content_condition(block):
                return self.execute_content_condition(block)
            else:
                return self.call_kb([block])

        if block.op == "AND":
            results = []

            content_filters, other_filters = [], []
            for item in block.items:
                if isinstance(item, ConditionBlock):
                    results.append(self.execute_blocks(item))
                else:
                    if self.is_content_condition(item):
                        content_filters.append(item)
                    else:
                        other_filters.append(item)
            if len(content_filters) > 0:
                content_filters2 = []
                exclude_ids = set()
                include_contents = set()
                # exclude content conditions
                for condition in content_filters:
                    ids = self.to_excluded_ids(condition, other_filters)
                    if ids is not None:
                        exclude_ids.update(ids)
                        continue
                    contents = self.to_include_content(condition)
                    if contents is not None:
                        include_contents.update(contents)
                        continue
                    else:
                        # keep origin content filter
                        content_filters2.append(condition)

                if exclude_ids:
                    # add to filter
                    values = [Constant(i) for i in exclude_ids]
                    condition = BinaryOperation(op="NOT IN", args=[Identifier(self.id_column), Tuple(values)])
                    other_filters.append(condition)
                # execute content filters
                if include_contents:
                    content = " AND ".join(include_contents)
                    result = self.execute_content_condition(
                        BinaryOperation(op="=", args=[Identifier(self.content_column), Constant(content)]),
                        other_filters,
                    )
                    results.append(result)
                for condition in content_filters2:
                    result = self.execute_content_condition(condition, other_filters)
                    results.append(result)
            elif len(other_filters) > 0:
                results.append(self.call_kb(other_filters))

            return self.intersect(results)

        elif block.op == "OR":
            results = []
            for item in block.items:
                results.append(self.execute_blocks(item))

            return self.union(results)

    def run(self, query: Select) -> pd.DataFrame:
        """
        Plan and execute query to KB. If query has complex conditions:
         - convert them to several queries with simple conditions, execute them and combine results

        Stages:
        - Remove unary NOT from condition: try to apply it to related operator
        - Flat conditions tree: convert into condition blocks:
           - having with same operators of the same levels in the same block
        - Recursively execute blocks
           - get data from OR blocks and union them
           - get data from AND blocks and intersect them

        :param query: select query
        :return: results
        """
        if query.where is not None:
            query_traversal(query.where, self.convert_unary_ops)
            blocks_tree = self.flatten_conditions(query.where)
            if query.limit is not None:
                self.limit = query.limit.value
            return self.execute_blocks(blocks_tree)
        else:
            return self.kb.select(query)
